import functools
import hashlib
from collections import deque
from inspect import signature
from typing import Callable

from egse.confman import ConfigurationManagerProxy
from egse.decorators import borg
from egse.env import bool_env
from egse.log import logger
from egse.response import Failure
from egse.response import Success

VERBOSE_DEBUG = bool_env("VERBOSE_DEBUG", default=False)


@borg
class ObservationContext:
    """Keep record of the observation context.

    Note: This class implements the Borg pattern and shares state between its instances.
          This allows us to keep track of the observation context across multiple building blocks and observations,
          without having to pass around an instance of this class.
    """

    def __init__(self) -> None:
        self.level = -1
        self.bbid_count = 0
        self.bbids = deque()

    def set_bbid(self, bbid: str):
        """Set the current building block id."""
        if bbid in self.bbids:
            raise ValueError(
                f"This building block ({bbid}) is already in execution, check dependencies between building blocks."
            )
        if self.level == -1:
            raise RuntimeError(f"This building block ({bbid}) is called outside the scope of an observation context.")

        self.bbids.append(bbid)
        self.bbid_count += 1
        self.level += 1

    def unset_bbid(self):
        """Unset the current building block id and go back to the previous level and bbid."""
        self.bbids.pop()
        self.level -= 1

    def get_current_bbid(self):
        return self.bbids[-1]

    def get_level(self):
        return self.level

    def start_observation(self, function_info: dict):
        if self.level == -1:
            self.level = 0
        else:
            raise Failure(
                "An observation can only be started when no observation is already running. "
                "Only one Observation can be active at a time."
            )

        try:
            with ConfigurationManagerProxy() as cm:
                rc = cm.start_observation(function_info)
                if not rc.successful:
                    self.level = -1
                    raise rc
                else:
                    return rc.return_code
        except ConnectionError as exc:
            self.level = -1
            raise Failure("Couldn't connect to the Configuration Manager Control Server", exc)

    def end_observation(self):
        if self.level > 0 or self.bbids:
            logger.warning(f"Observation contexts not empty at time of reset! (Level={self.level}, bbids={self.bbids})")
        self.level = -1
        self.bbid_count = 0
        self.bbids.clear()

        with ConfigurationManagerProxy() as cm:
            rc = cm.end_observation()
            if not rc.successful:
                raise rc

    def is_active(self):
        return self.level >= 0


def request_obsid():
    """Requests an `obsid` from the configuration manager.

    Returns:
        the current observation identifier.

    Raises:
        Failure with message and cause.
    """
    with ConfigurationManagerProxy() as cm:
        rc = cm.get_obsid()
        if isinstance(rc, Success):
            return rc.return_code

    raise rc


def stringify_args(args):
    s_args = []
    for arg in args:
        try:
            s_args.append(f"{arg.__module__}.{arg.__class__.__qualname__}")
        except AttributeError:
            s_args.append(repr(arg))

    return s_args


def stringify_kwargs(kwargs):
    s_kwargs = {}
    for k, v in kwargs.items():
        try:
            s_kwargs[k] = f"{v.__module__}.{v.__class__.__qualname__}"
        except AttributeError:
            s_kwargs[k] = repr(v)

    return s_kwargs


def execute(func: Callable, description=None, *args, **kwargs):
    """Execute a building block or observation."""

    try:
        ObservationContext().start_observation(
            {
                "func_name": func.__name__,
                "description": description,
                "args": stringify_args(args),
                "kwargs": stringify_kwargs(kwargs),
            }
        )

        obsid = request_obsid()
        logger.info(f"OBSID = {obsid}")

    except (Failure, TypeError) as exc:
        logger.error(f"Failed to start observation or test: {exc}")
        raise exc

    # can I check here if func is indeed a building_block?

    if not hasattr(func, "__building_block_func"):
        logger.warning("Executing a function that is not a building block.")

    try:
        response = func(*args, **kwargs)
    finally:
        ObservationContext().end_observation()

    return response


def start_observation(description: str):
    try:
        ObservationContext().start_observation({"description": description})

        obsid = request_obsid()
        logger.info(f"Observation started with obsid={obsid}")

    except Failure as exc:
        logger.error(f"Failed to start observation or test: {exc}")
        raise exc

    return obsid


def end_observation():
    ObservationContext().end_observation()


def building_block(func: Callable) -> Callable:
    """Mark a function as a building block and wrap it with lifecycle handling.

    The wrapper enforces keyword-only invocation, validates required keyword
    arguments against the function signature, sets the BBID on entry, and
    unsets it on exit (even if the function raises).

    Args:
        func: Function to decorate as a building block.

    Returns:
        A wrapped callable that preserves the original signature metadata.

    Raises:
        ValueError: If positional arguments are provided or required kwargs are missing.
    """

    setattr(func, "__building_block_func", True)

    def check_kwargs(**kwargs):
        sig = signature(func)
        missing = [par.name for par in sig.parameters.values() if par.name not in kwargs]

        if VERBOSE_DEBUG:
            logger.debug(f"missing keyword arguments: {missing}")

        if missing:
            raise ValueError(
                f"Expected {len(sig.parameters)} keyword parameters for "
                f"building block {func.__name__}, missing arguments are {missing}."
            )

    def rewrite_kwargs(**kwargs):
        if VERBOSE_DEBUG:
            logger.debug(f"{func.__name__}({kwargs})")
        return kwargs

    @functools.wraps(func)
    def wrapper_func(*args, **kwargs):
        # Take steps to make the function `func` into a building block.
        #
        # * set its building block id based on the name of the function
        # * check if this building block is executed within the scope of an observation
        # * execute the function `func`

        # TODO: We should also check for args, there can be no args!!

        if args:
            raise ValueError(
                f"Building block {func.__name__} can not have positional arguments. "
                f"The following arguments {args} should be given as keyword arguments."
            )

        kwargs = rewrite_kwargs(**kwargs)
        check_kwargs(**kwargs)

        start_building_block(func, *args, **kwargs)

        try:
            result = func(*args, **kwargs)
        finally:
            # Take steps to end the building block
            end_building_block()

        return result

    return wrapper_func


def start_building_block(func: Callable, *args, **kwargs) -> str:
    """Start a building block by assigning its BBID in the observation context.

    Args:
        func: Building block function used to derive the BBID.
        *args: Unused positional arguments (kept for compatibility).
        **kwargs: Unused keyword arguments (kept for compatibility).

    Returns:
        The generated BBID string.
    """

    # * define a unique building block identifier
    # * return the building block identifier

    bbid = get_bbid_uuid(func)
    ObservationContext().set_bbid(bbid)
    return bbid


def end_building_block():
    """End a building block by clearing the BBID from the observation context."""
    ObservationContext().unset_bbid()


def get_bbid_uuid(func: Callable) -> str:
    """Return a deterministic BBID for the given building block function.

    The BBID is derived from the function's fully-qualified name and is stable
    across runs. Only functions decorated with :func:`building_block` are
    accepted.

    Args:
        func: Building block function.

    Returns:
        A BBID string in the form ``BBID-<FUNCNAME>-<HASH>``.

    Raises:
        ValueError: If ``func`` is not callable or is not a building block.
    """
    if not isinstance(func, Callable):
        raise ValueError("The given argument should be a function or method.")

    if not hasattr(func, "__building_block_func"):
        raise ValueError("The given function is not defined as a building block.")

    func_identifier = f"{func.__module__}.{func.__qualname__}"
    unique_uuid = hashlib.sha256(func_identifier.encode()).hexdigest()
    bbid = f"BBID-{func.__name__.upper()}-{unique_uuid.upper()}"
    return bbid



import logging
import socket
from typing import Optional

import time

from egse.command import ClientServerCommand
from egse.device import DeviceConnectionInterface, DeviceTransport
from egse.settings import Settings

DEVICE_SETTINGS = Settings.load("LakeShore336 Controller")

logger = logging.getLogger(__name__)


class LakeShore336Command(ClientServerCommand):
    """Commands for the Lakeshore336 temperature controller.

    A Command is basically a string that is sent to a device and for which the device returns a response.  The command
    string can contain placeholders that will be filled when the command is called.  The arguments that are given, will
    be filled into the formatted string.  Arguments can be positional or keyword arguments, not both.
    """

    def get_cmd_string(self, *args, **kwargs) -> str:
        """Returns the formatted command string with the given positional and/or keyword arguments filled out.

        Args:
            *args: Positional arguments that are needed to construct the command string
            **kwargs: Keyword arguments that are needed to construct the command string
        """

        out = super().get_cmd_string(*args, **kwargs)
        return out + "\n"


class LakeShore336Error(Exception):
    """Base exception for all LakeShore336 errors."""

    pass


class LakeShore336EthernetInterface(DeviceConnectionInterface, DeviceTransport):
    """Defines the low-level interface to the LakeShore336 temperature controller."""

    def __init__(self, device_id: str, hostname: str = None, port: int = None):
        """Initialisation of an Ethernet interface for the Lakeshore336.

        Args:
            device_id (str): Device identifier
            hostname (str): Hostname to which to open a socket
            port (int): Port to which to open a socket
        """

        super().__init__()

        self.device_id = device_id

        self.hostname = DEVICE_SETTINGS[device_id]["HOSTNAME"] if hostname is None else hostname
        self.port = DEVICE_SETTINGS[device_id]["PORT"] if port is None else port

        self._socket = None
        self._is_connection_open = False

    def connect(self):
        """Connects to the Ethernet connection.

        Raises:
            LakeShore336Error when no connection could be established
        """

        # Sanity checks

        if self._is_connection_open:
            logger.warning(f"{self.device_id}: trying to connect to an already connected socket.")
            return

        if self.hostname in (None, ""):
            raise LakeShore336Error(f"{self.device_id}: hostname is not initialised.")

        if self.port in (None, 0):
            raise LakeShore336Error(f"{self.device_id}: port number is not initialised.")

        # Create a new socket instance

        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # self._socket.setblocking(1)
            # self._socket.settimeout(3)
        except socket.error as e_socket:
            raise LakeShore336Error(self.device_id, "Failed to create socket.") from e_socket

        # Attempt to establish a connection to the remote host

        # FIXME: Socket shall be closed on exception?

        # We set a timeout of 3s before connecting and reset to None (=blocking) after the `connect` method has been
        # called. This is because when no device is available, e.g. during testing, the timeout will take about
        # two minutes, which is way too long. It needs to be evaluated if this approach is acceptable and not causing
        # problems during production.

        try:
            logger.debug(f'Connecting a socket to host "{self.hostname}" using port {self.port}')
            self._socket.settimeout(3)
            self._socket.connect((self.hostname, self.port))
            self._socket.settimeout(None)
        except ConnectionRefusedError as exc:
            raise LakeShore336Error(self.device_id, f"Connection refused to {self.hostname}:{self.port}.") from exc
        except TimeoutError as exc:
            raise LakeShore336Error(self.device_id, f"Connection to {self.hostname}:{self.port} timed out.") from exc
        except socket.gaierror as exc:
            raise LakeShore336Error(self.device_id, f"Socket address info error for {self.hostname}") from exc
        except socket.herror as exc:
            raise LakeShore336Error(self.device_id, f"Socket host address error for {self.hostname}") from exc
        except socket.timeout as exc:
            raise LakeShore336Error(self.device_id, f"Socket timeout error for {self.hostname}:{self.port}") from exc
        except OSError as exc:
            raise LakeShore336Error(self.device_id, f"OSError caught ({exc}).") from exc

        self._is_connection_open = True

        # Check that we are connected to the controller by issuing the "*IDN?" or  query. If we don't get the right
        # response, then disconnect automatically.

        if not self.is_connected():
            raise LakeShore336Error(self.device_id, "Device is not connected, check logging messages for the cause.")

    def disconnect(self):
        """Disconnects from the Ethernet connection.

        Raises:
            LakeShore336Error when the socket could not be closed
        """

        try:
            if self._is_connection_open:
                logger.debug(f"Disconnecting from {self.hostname}")
                self._socket.close()
                self._is_connection_open = False
        except Exception as e_exc:
            raise LakeShore336Error(self.device_id, f"Could not close socket to {self.hostname}") from e_exc

    def reconnect(self):
        """Reconnects to the Ethernet connection."""

        if self._is_connection_open:
            self.disconnect()
        self.connect()

    def is_connected(self) -> bool:
        """Checks if the client is connected to the device.

        Returns: True if the client is connected to the device; False otherwise.
        """

        if not self._is_connection_open:
            return False

        try:
            version = self.query("*IDN?\n")
        except LakeShore336Error as exc:
            logger.exception(exc)
            logger.error("Most probably the client connection was closed. Disconnecting...")
            self.disconnect()
            return False

        if "LSCI" not in version:
            logger.error(f'Device did not respond correctly to a "*IDN?" command, response={version}. Disconnecting...')
            return False

        return True

    def write(self, command: str):
        """Sends a single command to the device controller without waiting for a response.

        Args
            command (str): Command to send to the controller
        """

        try:
            self._socket.sendall(command.encode())
        except socket.timeout as e_timeout:
            raise LakeShore336Error("Socket timeout error") from e_timeout
        except socket.error as e_socket:
            # Interpret any socket-related error as a connection error
            raise LakeShore336Error("Socket communication error.") from e_socket
        except AttributeError:
            if not self.is_connected:
                msg = "The LakeShore336 is not connected, use the connect() method."
                raise LakeShore336Error(msg)

    def trans(self, command: str) -> Optional[str]:
        """Sends a single command to the device and blocks until a response is received.

        This operation is seen as a transaction or query.

        Args:
            command (str): Command to send to the controller

        Returns: Either a string returned by the controller (on success) or an error message (on failure).

        Raises:
            LakeShore336Error when there was an I/O problem during communication with the controller, or when there
            was a timeout in either sending the command or receiving the response.
        """

        try:
            self._socket.sendall(command.encode())

            # wait for, read and return the response from HUBER (will be at most TBD chars)

            return_string = self.read()

            return return_string.decode().replace("\r\n", "").replace("+", "")

        except socket.timeout as e_timeout:
            raise LakeShore336Error(self.device_id, "Socket timeout error") from e_timeout
        except socket.error as e_socket:
            # Interpret any socket-related error as an I/O error
            raise LakeShore336Error(self.device_id, "Socket communication error.") from e_socket
        except ConnectionError as exc:
            raise LakeShore336Error(self.device_id, "Connection error.") from exc
        except AttributeError:
            if not self._is_connection_open:
                raise LakeShore336Error(self.device_id, "Device not connected, use the connect() method.")

    def read(self) -> bytes:
        """Reads the device buffer.

        Returns: Content of the device buffer.
        """

        n_total = 0
        buf_size = 2048

        try:
            for idx in range(100):
                time.sleep(0.05)  # Give the device time to fill the buffer
                data = self._socket.recv(buf_size)
                n = len(data)
                n_total += n
                if n < buf_size:
                    break
            return data
        except socket.timeout as e_timeout:
            logger.warning(f"Socket timeout error from {e_timeout}")
            return b"\r\n"



"""
A collection of useful decorator functions.
"""

import cProfile
import functools
import logging
import pstats
import time
import types
import warnings
from typing import Callable
from typing import List
from typing import Optional

import rich

from egse.log import logger
from egse.system import get_caller_info


def static_vars(**kwargs):
    """
    Define static variables in a function.

    The static variable can be accessed with <function name>.<variable name> inside the function body.

    Example:
        ```python
        @static_vars(count=0)
        def special_count():
            return special_count.count += 2
        ```

    """

    def decorator(func):
        for kw in kwargs:
            setattr(func, kw, kwargs[kw])
        return func

    return decorator


def implements_protocol(protocol):
    """
    Decorator to verify and document protocol compliance at class definition time.

    Usage:
        @implements_protocol(AsyncRegistryBackend)
        class MyBackend:
            ...
    """

    def decorator(cls):
        # Add protocol documentation
        if cls.__doc__:
            cls.__doc__ += f"\n\nThis class implements the {protocol.__name__} protocol."
        else:
            cls.__doc__ = f"This class implements the {protocol.__name__} protocol."

        # Store the protocol for reference
        cls.__implements_protocol__ = protocol

        # Add runtime verification method
        def _verify_protocol_compliance(self):
            if not isinstance(self, protocol):
                raise TypeError(f"{self.__class__.__name__} does not correctly implement {protocol.__name__}")
            return True

        cls.verify_protocol_compliance = _verify_protocol_compliance

        # Return the modified class
        return cls

    return decorator


def dynamic_interface(func) -> Callable:
    """Adds a static variable `__dynamic_interface` to a method.

    The intended use of this function is as a decorator for functions in an interface class.

    The static variable is currently used by the Proxy class to check if a method
    is meant to be overridden dynamically. The idea behind this is to loosen the contract
    of an abstract base class (ABC) into an interface. For an ABC, the abstract methods
    must be implemented at construction/initialization. This is not possible for the Proxy
    subclasses as they load their commands (i.e. methods) from the control server, and the
    method will be added to the Proxy interface after loading. Nevertheless, we like the
    interface already defined for auto-completion during development or interactive use.

    When a Proxy subclass that implements an interface with methods decorated by
    the `@dynamic_interface` does overwrite one or more of the decorated methods statically,
    these methods will not be dynamically overwritten when loading the interface from the
    control server. A warning will be logged instead.
    """
    setattr(func, "__dynamic_interface", True)
    return func


def query_command(func):
    """Adds a static variable `__query_command` to a method."""

    setattr(func, "__query_command", True)
    return func


def transaction_command(func):
    """Adds a static variable `__transaction_command` to a method."""

    setattr(func, "__transaction_command", True)
    return func


def read_command(func):
    """Adds a static variable `__read_command` to a method."""

    setattr(func, "__read_command", True)
    return func


def write_command(func):
    """Adds a static variable `__write_command` to a method."""

    setattr(func, "__write_command", True)
    return func


def average_time(*, name: str = "average_time", level: int = logging.INFO, precision: int = 6) -> Callable:
    """
    This is a decorator that is intended mainly as a development aid. When you decorate your function with
    `@average_time`, the execution time of your function will be kept and accumulated. At anytime in your code,
    you can request the total execution time and the number of calls:

        @average_time()
        def my_function():
            ...
        total_execution_time, call_count = my_function.report()

    Requesting the report will automatically log the average runtime and the number of calls.
    If you need to reset the execution time and the number of calls during your testing, use:

        my_function.reset()

    Args:
        name: A name for the timer that will be used during reporting, default='average_time'
        level: the required log level, default=logging.INFO
        precision: the precision used to report the average time, default=6

    Returns:
        The decorated function.

    """

    def actual_decorator(func):
        func._run_time = 0.0
        func._call_count = 0

        def _report_average_time():
            if func._call_count:
                average_time = func._run_time / func._call_count
                logger.log(
                    level,
                    f"{name}: "
                    f"average runtime of {func.__name__!r} is {average_time:.{precision}f}s, "
                    f"#calls = {func._call_count}.",
                )
            else:
                logger.log(level, f"{name}: function {func.__name__!r} was never called.")

            return func._run_time, func._call_count

        def _reset():
            func._run_time = 0
            func._call_count = 0

        func.report = _report_average_time
        func.reset = _reset

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = func(*args, **kwargs)
            end_time = time.perf_counter()
            func._run_time += end_time - start_time
            func._call_count += 1
            return result

        return wrapper

    return actual_decorator


def timer(*, name: str = "timer", level: int = logging.INFO, precision: int = 4):
    """
    Print the runtime of the decorated function.

    Args:
        name: a name for the Timer, will be printed in the logging message
        level: the logging level for the time message [default=INFO]
        precision: the number of decimals for the time [default=3 (ms)]
    """

    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper_timer(*args, **kwargs):
            start_time = time.perf_counter()
            value = func(*args, **kwargs)
            end_time = time.perf_counter()
            run_time = end_time - start_time
            logger.log(level, f"{name}: Finished {func.__name__!r} in {run_time:.{precision}f} secs")
            return value

        return wrapper_timer

    return actual_decorator


def async_timer(*, name: str = "timer", level: int = logging.INFO, precision: int = 4):
    """
    Print the runtime of the decorated async function.

    Args:
        name: a name for the Timer, will be printed in the logging message
        level: the logging level for the time message [default=INFO]
        precision: the number of decimals for the time [default=3 (ms)]
    """

    def actual_decorator(func):
        @functools.wraps(func)
        async def wrapper_timer(*args, **kwargs):
            start_time = time.perf_counter()
            value = await func(*args, **kwargs)
            end_time = time.perf_counter()
            run_time = end_time - start_time
            logger.log(level, f"{name}: Finished {func.__name__!r} in {run_time:.{precision}f} secs")
            return value

        return wrapper_timer

    return actual_decorator


def time_it(count: int = 1000, precision: int = 4) -> Callable:
    """Print the runtime of the decorated function.

    This is a simple replacement for the builtin ``timeit`` function. The purpose is to simplify
    calling a function with some parameters.

    The intended way to call this is as a function:

        value = function(args)

        value = time_it(10_000)(function)(args)

    The `time_it` function can be called as a decorator in which case it will always call the
    function `count` times which is probably not what you want.

    Args:
        count (int): the number of executions [default=1000].
        precision (int): the number of significant digits [default=4]

    Returns:
        value: the return value of the last function execution.

    See also:
        the ``Timer`` context manager located in ``egse.system``.

    Usage:
        ```python
        @time_it(count=10000)
        def function(args):
            pass

        time_it(10000)(function)(args)
        ```
    """

    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper_timer(*args, **kwargs):
            value = None
            start_time = time.perf_counter()
            for _ in range(count):
                value = func(*args, **kwargs)
            end_time = time.perf_counter()
            run_time = end_time - start_time
            logging.info(
                f"Finished {func.__name__!r} in {run_time / count:.{precision}f} secs "
                f"(total time: {run_time:.2f}s, count: {count})"
            )
            return value

        return wrapper_timer

    return actual_decorator


def debug(func):
    """Logs the function signature and return value."""

    @functools.wraps(func)
    def wrapper_debug(*args, **kwargs):
        if __debug__:
            args_repr = [repr(a) for a in args]
            kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            logger.debug(f"Calling {func.__name__}({signature})")
            value = func(*args, **kwargs)
            logger.debug(f"{func.__name__!r} returned {value!r}")
        else:
            value = func(*args, **kwargs)
        return value

    return wrapper_debug


def profile_func(
    output_file: str | None = None,
    sort_by: str = "cumulative",
    lines_to_print: int | None = None,
    strip_dirs: bool = False,
) -> Callable:
    """A time profiler decorator.

    Args:
        output_file: str or None. Default is None
            Path of the output file. If only name of the file is given, it's
            saved in the current directory.
            If it's None, the name of the decorated function is used.
        sort_by: str or SortKey enum or tuple/list of str/SortKey enum
            Sorting criteria for the Stats object.
            For a list of valid string and SortKey refer to:
            https://docs.python.org/3/library/profile.html#pstats.Stats.sort_stats
        lines_to_print: int or None
            Number of lines to print. Default (None) is for all the lines.
            This is useful in reducing the size of the printout, especially
            that sorting by 'cumulative', the time consuming operations
            are printed toward the top of the file.
        strip_dirs: bool
            Whether to remove the leading path info from file names.
            This is also useful in reducing the size of the printout

    Returns:
        Profile of the decorated function

    Note:
        This code was taken from this gist: [a profile
        decorator](https://gist.github.com/ekhoda/2de44cf60d29ce24ad29758ce8635b78).

        Inspired by and modified the profile decorator of Giampaolo Rodola:
        [profile decorator](http://code.activestate.com/recipes/577817-profile-decorator/).


    """

    def inner(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _output_file = output_file or func.__name__ + ".prof"
            pr = cProfile.Profile()
            pr.enable()
            retval = func(*args, **kwargs)
            pr.disable()
            pr.dump_stats(_output_file)

            with open(_output_file, "w") as f:
                ps = pstats.Stats(pr, stream=f)
                if strip_dirs:
                    ps.strip_dirs()
                if isinstance(sort_by, (tuple, list)):
                    ps.sort_stats(*sort_by)
                else:
                    ps.sort_stats(sort_by)
                ps.print_stats(lines_to_print)
            return retval

        return wrapper

    return inner


def profile(func):
    """
    Prints the function signature and return value to stdout.

    This function checks the `Settings.profiling()` value and only prints out
    profiling information if this returns True.

    Profiling can be activated with `Settings.set_profiling(True)`.
    """
    if not hasattr(profile, "counter"):
        profile.counter = 0

    from egse.settings import Settings

    @functools.wraps(func)
    def wrapper_profile(*args, **kwargs):
        if Settings.profiling():
            profile.counter += 1
            args_repr = [repr(a) for a in args]
            kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            caller = get_caller_info(level=2)
            prefix = f"PROFILE[{profile.counter}]: "
            rich.print(f"{prefix}Calling {func.__name__}({signature})")
            rich.print(f"{prefix}    from {caller.filename} at {caller.lineno}.")
            value = func(*args, **kwargs)
            rich.print(f"{prefix}{func.__name__!r} returned {value!r}")
            profile.counter -= 1
        else:
            value = func(*args, **kwargs)
        return value

    return wrapper_profile


class Profiler:
    """
    A simple profiler class that provides some useful functions to profile a function.

    - count: count the number of times this function is executed
    - duration: measure the total and average duration of the function [seconds]

    Examples:
          >>> from egse.decorators import Profiler
          >>> @Profiler.count()
          ... def square(x):
          ...     return x**2

          >>> x = [square(x) for x in range(1_000_000)]

          >>> print(f"Function 'square' called {square.get_count()} times.")
          >>> print(square)

          >>> @Profiler.duration()
          ... def square(x):
          ...     time.sleep(0.1)
          ...     return x**2

          >>> x = [square(x) for x in range(100)]

          >>> print(f"Function 'square' takes on average {square.get_average_duration():.6f} seconds.")
          >>> print(square)

    """

    class CountCalls:
        def __init__(self, func):
            self.func = func
            self.count = 0

        def __call__(self, *args, **kwargs):
            self.count += 1
            return self.func(*args, **kwargs)

        def get_count(self):
            return self.count

        def reset(self):
            self.count = 0

        def __str__(self):
            return f"Function '{self.func.__name__}' was called {self.count} times."

        # The __get__ method is here to make the decorator work with instance methods (methods inside a class)
        # as well. It ensures that when the decorated method is called on an instance, the self argument is
        # correctly passed to the method.

        def __get__(self, instance, owner):
            if instance is None:
                return self
            else:
                return types.MethodType(self, instance)

    class Duration:
        def __init__(self, func):
            self.func = func
            self.duration = 0
            self.count = 0

        def __call__(self, *args, **kwargs):
            start = time.perf_counter_ns()
            response = self.func(*args, **kwargs)
            self.count += 1
            self.duration += time.perf_counter_ns() - start
            return response

        def get_count(self):
            return self.count

        def get_duration(self):
            return self.duration / 1_000_000_000

        def get_average_duration(self):
            return self.duration / 1_000_000_000 / self.count if self.count else 0.0

        def reset(self):
            self.duration = 0
            self.count = 0

        def __str__(self):
            return f"Function '{self.func.__name__}' takes on average {self.get_average_duration():.6f} seconds."

        # The __get__ method is here to make the decorator work with instance methods (methods inside a class)
        # as well. It ensures that when the decorated method is called on an instance, the self argument is
        # correctly passed to the method.

        def __get__(self, instance, owner):
            if instance is None:
                return self
            else:
                return types.MethodType(self, instance)

    @classmethod
    def count(cls):
        return cls.CountCalls

    @classmethod
    def duration(cls):
        return cls.Duration


def to_be_implemented(func):
    """Print a warning message that this function/method has to be implemented."""

    @functools.wraps(func)
    def wrapper_tbi(*args, **kwargs):
        logger.warning(f"The function/method {func.__name__} is not yet implemented.")
        return func(*args, **kwargs)

    return wrapper_tbi


# Taken and adapted from https://github.com/QCoDeS/Qcodes


def deprecate(reason: Optional[str] = None, alternative: Optional[str] = None) -> Callable:
    """
    Deprecate a function or method. This will print a warning with the function name and where
    it is called from. If the optional parameters `reason` and `alternative` are given, that
    information will be printed with the warning.

    Examples:

        @deprecate(reason="it doesn't follow PEP8", alternative="set_color()")
        def setColor(self, color):
            self.set_color(color)

    Args:
        reason: provide a short explanation why this function is deprecated. Generates 'because {reason}'
        alternative: provides an alternative function/parameters to be used. Generates 'Use {alternative}
            as an alternative'

    Returns:
        The decorated function.
    """

    def actual_decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def decorated_func(*args, **kwargs):
            caller = get_caller_info(2)
            msg = f'The function "{func.__name__}" used at {caller.filename}:{caller.lineno} is deprecated'
            if reason is not None:
                msg += f", because {reason}"
            if alternative is not None:
                msg += f". Use {alternative} as an alternative"
            msg += "."
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        decorated_func.__doc__ = (
            f"This function is DEPRECATED, because {reason}, use {alternative} as an alternative.\n"
        )
        return decorated_func

    return actual_decorator


def singleton(cls):
    """
    Use class as a singleton.

    from:
        [Decorator library: Signleton](https://wiki.python.org/moin/PythonDecoratorLibrary#Singleton)
    """

    cls.__new_original__ = cls.__new__

    @functools.wraps(cls.__new__)
    def singleton_new(cls, *args, **kw):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it

        cls.__it__ = it = cls.__new_original__(cls, *args, **kw)
        it.__init_original__(*args, **kw)
        return it

    cls.__new__ = singleton_new
    cls.__init_original__ = cls.__init__
    cls.__init__ = object.__init__

    return cls


def borg(cls):
    """
    Use the Borg pattern to make a class with a shared state between its instances and subclasses.

    from:
        [we don't need no singleton](
        http://code.activestate.com/recipes/66531-singleton-we-dont-need-no-stinkin-singleton-the-bo/)
    """

    cls._shared_state = {}
    orig_init = cls.__init__

    def new_init(self, *args, **kwargs):
        self.__dict__ = cls._shared_state
        orig_init(self, *args, **kwargs)

    cls.__init__ = new_init

    return cls


class classproperty:
    """Defines a read-only class property.

    Examples:

        >>> class Message:
        ...     def __init__(self, msg):
        ...         self._msg = msg
        ...
        ...     @classproperty
        ...     def name(cls):
        ...         return cls.__name__

        >>> msg = Message("a simple doctest")
        >>> assert "Message" == msg.name

    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        return self.func(owner)

    def __set__(self, instance, value):
        raise AttributeError(
            f"Cannot change class property '{self.func.__name__}' for class '{instance.__class__.__name__}'."
        )


class Nothing:
    """Just to get a nice repr for Nothing. It is kind of a Null object..."""

    def __repr__(self):
        return "<Nothing>"


def spy_on_attr_change(obj: object, obj_name: str = None) -> None:
    """
    Tweak an object to show attributes changing. The changes are reported as WARNING log messages
    in the `egse.spy` logger.

    Note this is not a decorator, but a function that changes the class of an object.

    Note that this function is a debugging aid and should not be used in production code!

    Args:
        obj (object): any object that you want to monitor
        obj_name (str): the variable name of the object that was given in the code, if None than
            the class name will be printed.

    Example:
        ```python
        class X:
           pass

        x = X()
        spy_on_attr_change(x, obj_name="x")
        x.a = 5
        ```

    From:
        [Adding a dunder to an object](https://nedbatchelder.com/blog/202206/adding_a_dunder_to_an_object.html)
    """
    logger = logging.getLogger("egse.spy")

    class Wrapper(obj.__class__):
        def __setattr__(self, name, value):
            old = getattr(self, name, Nothing())
            logger.warning(f"Spy: in {obj_name or obj.__class__.__name__} -> {name}: {old!r} -> {value!r}")
            return super().__setattr__(name, value)

    class_name = obj.__class__.__name__
    obj.__class__ = Wrapper
    obj.__class__.__name__ = class_name


def retry_with_exponential_backoff(
    max_attempts: int = 5, initial_wait: float = 1.0, backoff_factor: int = 2, exceptions: List = None
) -> Callable:
    """
    Decorator for retrying a function with exponential backoff.

    This decorator can be applied to a function to handle specified exceptions by
    retrying the function execution. It will make up to 'max_attempts' attempts with a
    waiting period that grows exponentially between each attempt (dependent on the backoff_factor).
    Any exception from the list provided in the `exceptions` argument will be ignored for the
    given `max_attempts`.

    If after all attempts still an exception is raised, it will be passed through the
    calling function, otherwise the functions return value will be returned.

    Args:
        max_attempts: The maximum number of attempts to make.
        initial_wait: The initial waiting time in seconds before retrying after the first failure.
        backoff_factor: The factor by which the wait time increases after each failure.
        exceptions: list of exceptions to ignore, if None all exceptions will be ignored `max_attempts`.

    Returns:
        The response from the executed function.
    """

    exceptions = [Exception] if exceptions is None else exceptions

    def actual_decorator(func):
        @functools.wraps(func)
        def decorate_func(*args, **kwargs):
            attempt = 0
            wait_time = initial_wait
            last_exception = None

            while attempt < max_attempts:
                try:
                    response = func(*args, **kwargs)  # Attempt to call the function
                    logger.info(f"{func.__name__} successfully executed.")
                    return response
                except tuple(exceptions) as exc:
                    last_exception = exc
                    attempt += 1
                    logger.info(
                        f"Retry {attempt}: {func.__name__} will be executing again in {wait_time * backoff_factor}s. "
                        f"Received a {last_exception!r}."
                    )
                    time.sleep(wait_time)  # Wait before retrying
                    wait_time *= backoff_factor  # Increase wait time for the next attempt

            # If the loop completes, all attempts have failed, reraise the last exception

            raise last_exception

        return decorate_func

    return actual_decorator


def retry(times: int = 3, wait: float = 10.0, exceptions: List = None) -> Callable:
    """
    Decorator that retries a function multiple times with a delay between attempts.

    This decorator can be applied to a function to handle specified exceptions by
    retrying the function execution. It will make up to 'times' attempts with a
    waiting period of 'wait' seconds between each attempt. Any exception from the
    list provided in the `exceptions` argument will be ignored for the given `times`.

    If after times attempts still an exception is raised, it will be passed through the
    calling function, otherwise the functions return value will be returned.

    Args:
        times (int, optional): The number of retry attempts. Defaults to 3.
        wait (float, optional): The waiting period between retries in seconds. Defaults to 10.0.
        exceptions (List[Exception] or None, optional): List of exception types to catch and retry.
            Defaults to None, which catches all exceptions.

    Returns:
        Callable: The decorated function.

    Example:
        Apply the retry decorator to a function with specific retry settings:

        ```python
        @retry(times=5, wait=15.0, exceptions=[ConnectionError, TimeoutError])
        def my_function():
            # Function logic here
        ```

    Note:
        The decorator catches specified exceptions and retries the function, logging
        information about each retry attempt.

    """

    exceptions = [Exception] if exceptions is None else exceptions

    def actual_decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def decorated_func(*args, **kwargs):
            previous_exception = None
            for n in range(times):
                try:
                    return func(*args, **kwargs)
                except tuple(exceptions) as exc:
                    previous_exception = exc
                if n < times:
                    logger.info(
                        f"Retry {n + 1}: {func.__name__} will be executing again in {wait}s. "
                        f"Received a {previous_exception!r}."
                    )
                    time.sleep(wait)
            raise previous_exception

        return decorated_func

    return actual_decorator


def execution_count(func):
    """Counts the number of times the function has been executed."""
    func._call_count = 0

    def counts():
        return func._call_count

    def reset():
        func._call_count = 0

    func.counts = counts
    func.reset = reset

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        func._call_count += 1
        value = func(*args, **kwargs)
        return value

    return wrapper



class Car:
    __match_args__ = ('key', 'name')

    def __init__(self, key, name):
        self.key = key
        self.name = name


def main():
    expr = eval(input('Expr: '))
    match expr:
        case (0, x):              # seq of 2 elems with first 0
            print(f'(0, {x})')    # (new variable x set to second elem)
        case ['a', x, 'c']:       # seq of 3 elems: 'a', anything, 'c'
            print(f"'a', {x!r}, 'c'")
        case {'foo': bar}:        # dict with key 'foo' (may have others)
            print(f"{{'foo': {bar}}}")
        case [1, 2, *rest]:       # seq of: 1, 2, ... other elements
            print(f'[1, 2, *{rest}]')
        case {'x': x, **kw}:      # dict with key 'x' (others go to kw)
            print(f"{{'x': {x}, **{kw}}}")
        case Car(key=key, name='Tesla'):  # Car with name 'Tesla' (any key)
            print(f"Car({key!r}, 'TESLA!')")
        case Car(key, name):      # similar to above, but use __match_args__
            print(f"Car({key!r}, {name!r})")
        case 1 | 'one' | 'I':     # int 1 or str 'one' or 'I'
            print('one')
        case ['a'|'b' as ab, c]:  # seq of 2 elems with first 'a' or 'b'
            print(f'{ab!r}, {c!r}')
        case (x, y) if x == y:    # seq of 2 elems with first equal to second
            print(f'({x}, {y}) with x==y')
        case _:
            print('no match')
            return "Quit"


if __name__ == '__main__':

    while True:
        if main() == "Quit":
            break



"""Ariel Telescope Control Unit (TCU) commanding.

Commanding is done by communicating with the TCU Arduino via a serial port (with the PySerial package).

We discern between the following types of commands:
    - General commands;
    - M2MD commands (on a per-axis basis);
    - Thermal Monitoring System (TSM) commands;
    - HK commands.

Reference documents:
    - RD01: TCU User Manual (ARIEL-IEEC-PL-TN-002), v1.2
    - RD02: ARIEL TCU Data Handling (ARIEL-IEEC-PL-TN-007), v1.0
    - RD03: TCU code provided by Vladimiro Noce (priv. comm.)
    - RD04: ARIEL Telescope Control Unit Design Description Document (ARIEL-IEEC-PL-DD-001), v1.10
    - RD05: ARIEL TCU FW Architecture Design (ARIEL-IEEC-PL-DD-002), v1.5
"""

import logging

from serial.tools import list_ports

from egse.ariel.tcu import PROXY_TIMEOUT, SERVICE_TYPE, TcuMode, PROTOCOL, HOSTNAME, COMMANDING_PORT
from egse.ariel.tcu.tcu_cmd_utils import general_cmd, m2md_cmd, tsm_cmd, hk_cmd
from egse.ariel.tcu.tcu_cmd_utils import (
    set_tcu_mode,
    tcu_simulated,
    restart_links_period_latch,
    set_restart_links_period,
    CommandAddress,
    ope_mng_command,
    ope_mng_event_clear_protect_flag,
    ope_mng_event_clear,
    ope_mng_status,
    ope_mng_event_reg,
    get_acq_curr_off_corr,
    set_acq_curr_off_corr,
    get_acq_curr_gain_corr,
    set_acq_curr_gain_corr,
    acq_axis_a_curr_read,
    acq_axis_b_curr_read,
    acq_ave_lpf_en,
    acq_ovc_cfg_filter,
    acq_avc_filt_time,
    acq_average_type,
    acq_spk_filt_counter_lim,
    acq_spk_filt_incr_thr,
    get_prof_gen_axis_step,
    set_prof_gen_axis_step,
    get_prof_gen_axis_speed,
    set_prof_gen_axis_speed,
    get_prof_gen_axis_state_start,
    set_prof_gen_axis_state_start,
    sw_rs_xx_sw_rise,
    sw_rs_xx_sw_fall,
    set_tsm_current_value,
    set_tsm_current_offset,
    set_tsm_adc_hpf_register,
    set_tsm_adc_ofc_register,
    set_tsm_adc_fsc_register,
    tsm_adc_command,
    tsm_adc_calibration,
    tsm_adc_value_xx_currentn,
    tsm_adc_value_xx_biasn,
    tsm_adc_value_xx_currentp,
    tsm_adc_value_xx_biasp,
    tcu_firmware_id,
    get_tcu_mode,
    tcu_status,
    get_restart_links_period,
    tsm_latch,
    get_tsm_current_value,
    get_tsm_current_offset,
    tsm_adc_id_register,
    tsm_adc_configuration_register,
    get_tsm_adc_ofc_register,
    get_tsm_adc_fsc_register,
    tsm_adc_command_latch,
    tsm_acq_counter,
    get_tsm_adc_hpf_register,
    tsm_adc_register_latch,
    vhk_psu_vmotor,
    vhk_psu_vhi,
    vhk_psu_vlow,
    vhk_psu_vmedp,
    vhk_psu_vmedn,
    ihk_psu_vmedn,
    ihk_psu_vlow,
    ihk_psu_vmedp,
    ihk_psu_vhi,
    ihk_psu_vmotor,
    thk_psu_first,
    thk_m2md_first,
    thk_psu_second,
    thk_m2md_second,
    thk_cts_q1,
    thk_cts_q2,
    thk_cts_q3,
    thk_cts_q4,
    thk_cts_fpga,
    thk_cts_ads1282,
    vhk_ths_ret,
    hk_acq_counter,
)
from egse.ariel.tcu.tcu_devif import TcuDeviceInterface, TcuHexInterface
from egse.device import DeviceInterface
from egse.mixin import dynamic_command, CommandType, DynamicCommandMixin
from egse.proxy import DynamicProxy
from egse.registry.client import RegistryClient
from egse.zmq_ser import connect_address

LOGGER = logging.getLogger("egse.ariel.tcu")


def get_all_serial_ports() -> list:
    """Returns a list of all available serial ports.

    Returns:
        List of all available serial ports.
    """

    ports = list_ports.comports()

    for port in ports:
        print(port)

    return ports


class TcuInterface(DeviceInterface):
    # General commands

    @general_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tcu_firmware_id)
    def tcu_firmware_id(self):
        """Selects the Instrument Control Unit (ICU) channel and returns the firmware version.

        Returns:
            Firmware version of the Ariel TCU.
        """

        pass

    @general_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_tcu_mode)
    def get_tcu_mode(self):
        """Returns the current mode of the Ariel TCU.

        Possible modes are:
            - IDLE (0x0000): Waiting for commands, minimum power consumption
            - BASE (0x0001): HK + TSM Circuitry on
            - CALIBRATION (0x0003): HK + TSM + M2MD circuitry on

        Returns:
            Current mode of the Ariel TCU.
        """

        pass

    @general_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_tcu_mode)
    def set_tcu_mode(self, tcu_mode: TcuMode | int = TcuMode.IDLE):
        """Selects the Ariel TCU working mode.

        Args:
            tcu_mode (TcuMode | int): Ariel TCU working mode:
                - IDLE (0x0000): Waiting for commands, minimum power consumption
                - BASE (0x0001): HK + TSM Circuitry on
                - CALIBRATION (0x0003): HK + TSM + M2MD circuitry on
        """

        pass

    @general_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tcu_status)
    def tcu_status(self):
        """Returns the TCU status.

        The TCU status is a bit word that indicates the current state of the TCU.  The meaning of a bit being set
        to one is as follows:

            - bit 0: main link enabled
            - bit 1: secondary link enabled
            - bit 2: TSM enabled
            - bit3: M2MD enabled
            - bit 4: TSM in simulated mode
            - bit 5: M2MD in simulated mode
            - bit 6: HK in simulated mode
            - bit 7: M2MD axis 1 enabled
            - bit 8: M2MD axis 2 enabled
            - bit 9: M2MD axis 3 enabled
            - bit 10: TSM initialised
            - bit 11: ADS_MFLAG active
            - bit 12: ADS_REGISTERS_ERROR active

        Returns:
            Bit word that indicated the current state of the TCU.
        """

        pass

    @general_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tcu_simulated)
    def tcu_simulated(self, cargo2: int):
        """Changes a TCU sub-system in simulated mode.

        This is only possible in IDLE mode.

        The cargo2 parameter denotes which sub-system to change to simulated mode.  The meaning of a bit being set to
        one is as follows:

            - bit 0: TBD
            - bit 1: TBD
            - bit 2: TBD
            - bit 3: TBD
            - bit 4: Put the TSM in simulated mode
            - bit 5: Put the M2MD in simulated mode
            - bit 6: Put the HK in simulated mode

        Args:
            cargo2 (int): Cargo 2 part of the command string.
        """
        pass

    @general_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=restart_links_period_latch)
    def restart_links_period_latch(self, cargo2: int):
        """Re-starts the link period latch.

        This is only possible in IDLE mode.


        Note that a read does not disarm the latch.

        Args:
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @general_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_restart_links_period)
    def get_restart_links_period(self):
        pass

    @general_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_restart_links_period)
    def set_restart_links_period(self, link_period: int = 0xFFFF):
        """Re-start both links if no message is received after the given link period +1s.

        Args:
            link_period (int): 1s after this time duration, both links will be re-started if no message is received.
        """
        pass

    # M2MD commands

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ope_mng_command)
    def ope_mng_command(self, axis: CommandAddress | str | int, cargo2: int = 0x0002):
        """Commands the action to the SENER motor driver IP core for the given M2MD axis.

        The cargo2 parameter denotes the action to be performed.  The meaning of a bit being set to
        one is as follows:

            - bit 0: Activate motion
            - bit 1: Stop motion

        Args:
            axis (CommandAddress): Axis to which the command is sent.
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ope_mng_event_clear_protect_flag)
    def ope_mng_event_clear_protect_flag(self, axis: CommandAddress | str | int, cargo2: int = 0xAAAA):
        """Clears the event register protection flag.

        Args:
            axis (CommandAddress): Axis to which the command is sent.
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ope_mng_event_clear)
    def ope_mng_event_clear(self, axis: CommandAddress | str | int, cargo2: int = 0x0001):
        """Clears the event register for the given M2MD axis.

        Args:
            axis (CommandAddress): Axis to which the command is sent.
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ope_mng_status)
    def ope_mng_status(self, axis: CommandAddress | str | int):
        """Returns the current status of the motor for the given M2MD axis.

        Args:
            axis (CommandAddress): Axis to which the command is sent.

        Returns:
            Current status of the motor for the given M2MD axis. Bit 0 being set to one means that the motor is
            in stand-by mode; bit 1 being set to one means that the motor is in operation state.
        """

        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ope_mng_event_reg)
    def ope_mng_event_reg(self, axis: CommandAddress | str | int):
        """Returns the list of all events since wake-up or the last clear event.

        Args:
            axis (CommandAddress): Axis to which the command is sent.

        Returns:
            List of all events since wake-up or the last clear event.
        """

        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_acq_curr_off_corr)
    def get_acq_curr_off_corr(self, axis: CommandAddress | str | int):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_acq_curr_off_corr)
    def set_acq_curr_off_corr(self, axis: CommandAddress | str | int, cargo2: int = 0x03FB):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_acq_curr_gain_corr)
    def get_acq_curr_gain_corr(self, axis: CommandAddress | str | int):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_acq_curr_gain_corr)
    def set_acq_curr_gain_corr(self, axis: CommandAddress | str | int, cargo2: int = 0x074C):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=acq_axis_a_curr_read)
    def acq_axis_a_curr_read(self, axis: CommandAddress | str | int):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=acq_axis_b_curr_read)
    def acq_axis_b_curr_read(self, axis: CommandAddress | str | int):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=acq_ave_lpf_en)
    def acq_ave_lpf_en(self, axis: CommandAddress | str | int, cargo2: int = 0x0001):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=acq_ovc_cfg_filter)
    def acq_ovc_cfg_filter(self, axis: CommandAddress | str | int, cargo2: int = 0):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=acq_avc_filt_time)
    def acq_avc_filt_time(self, axis: CommandAddress | str | int, cargo2: int = 0):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=acq_average_type)
    def acq_average_type(self, axis: CommandAddress | str | int, cargo2: int = 0x0000):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=acq_spk_filt_counter_lim)
    def acq_spk_filt_counter_lim(self, axis: CommandAddress | str | int, cargo2: int = 0x0001):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=acq_spk_filt_incr_thr)
    def acq_spk_filt_incr_thr(self, axis: CommandAddress | str | int, cargo2: int = 0x04C0):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_prof_gen_axis_step)
    def get_prof_gen_axis_step(self, axis: CommandAddress | str | int):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_prof_gen_axis_step)
    def set_prof_gen_axis_step(self, axis: CommandAddress | str | int, cargo2: int = 0x0480):
        """Axis position command for the given M2MD axis.

        The cargo2 parameter denotes the desired positioning. The meaning of its constituent bits is as follows:

            - bits 0-14: Number of steps to be carried out
            - bit 15: Direction of motion (0: counterclockwise, 1: clockwise)

        Args:
            axis (CommandAddress | str | int): Axis to which the command is sent.
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_prof_gen_axis_speed)
    def get_prof_gen_axis_speed(self, axis: CommandAddress | str | int):
        """Returns the axis writing speed for the given M2MD axis.

        Args:
            axis (CommandAddress | str | int): Axis to which the command is sent.
        """
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_prof_gen_axis_speed)
    def set_prof_gen_axis_speed(self, axis: CommandAddress | str | int, speed: int = 0x0177):
        """Axis velocity command for the given M2MD axis.

        The cargo2 parameter denotes the desired velocity.

        Args:
            axis (CommandAddress | str | int): Axis to which the command is sent.
            speed (int): Cargo 2 part of the command string.
        """

        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_prof_gen_axis_state_start)
    def get_prof_gen_axis_state_start(self, axis: CommandAddress | str | int):
        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_prof_gen_axis_state_start)
    def set_prof_gen_axis_state_start(self, axis: CommandAddress | str | int, cargo2: int = 0):
        """Changes the starting point of the magnetic state for the given M2MD axis.


        Args:
            axis (CommandAddress | str | int): Axis to which the command is sent.
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=sw_rs_xx_sw_rise)
    def sw_rs_xx_sw_rise(self, axis: CommandAddress | str | int, position: int = 1):
        """Position switch rise.

        Args:
            axis (CommandAddress | str | int): Axis to which the command is sent.
            position (int): Position of the SW_RS_XX_SW_RISE command.

        Returns:
            Relative position measured when the rising edge of the given switch is detected.
        """

        pass

    @m2md_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=sw_rs_xx_sw_fall)
    def sw_rs_xx_sw_fall(self, axis: CommandAddress | str | int, position: int = 1):
        """Position switch fall.

        Args:
            axis (CommandAddress | str | int): Axis to which the command is sent.
            position (int): Position of the SW_RS_XX_SW_FALL command.

        Returns:
            Relative position measured when the falling edge of the given switch is detected.
        """

        pass

    # TSM commands

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_latch)
    def tsm_latch(self, cargo1: str | int, cargo2: int = 0):
        """Latches to allow the modification of the operation management register.

        Args:
            cargo1 (str | int): Cargo 1 part of the command string.
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_tsm_current_value)
    def get_tsm_current_value(self):
        """Returns the TSM current.

        Returns:
            TSM current.
        """
        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_tsm_current_value)
    def set_tsm_current_value(self, cargo1: int = 0, cargo2: int = 0):
        """Sets the TSM current value.

        Args:
            cargo1 (int): Cargo 1 part of the command string.
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_tsm_current_offset)
    def get_tsm_current_offset(self):
        """Returns the TSM current offset.

        Returns:
            TSM current offset.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_tsm_current_offset)
    def set_tsm_current_offset(self, cargo1: int = 0, cargo2: int = 0):
        """Sets the TSM current offset.

        Args:
            cargo1 (int): Cargo 1 part of the command string.
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_register_latch)
    def tsm_adc_register_latch(self, cargo1: int = 0, cargo2: int = 0):
        """Re-starts the TSM ADC register latch.

        Args:
            cargo1 (int): Cargo 1 part of the command string.
            cargo2 (int): Cargo 2 part of the command string.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_id_register)
    def tsm_adc_id_register(self):
        """Returns the content of the TSM ADC identifier register.

        Returns:
            Content of the TSM ADC identifier register.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_configuration_register)
    def tsm_adc_configuration_register(self):
        """Returns the content of the TSM ADC configuration register.

        Returns:
            Content of the TSM ADC configuration register.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_tsm_adc_hpf_register)
    def get_tsm_adc_hpf_register(self):
        """Returns the content of the high-pass corner frequency register.

        Returns:
            Content of the high-pass corner frequency register.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_tsm_adc_hpf_register)
    def set_tsm_adc_hpf_register(self, cargo1: int = 0, cargo2: int = 0):
        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_tsm_adc_ofc_register)
    def get_tsm_adc_ofc_register(self):
        """Returns the content of the offset calibration register.

        Returns:
            Content of the offset calibration register.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_tsm_adc_ofc_register)
    def set_tsm_adc_ofc_register(self, cargo1: int = 0, cargo2: int = 0):
        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=get_tsm_adc_fsc_register)
    def get_tsm_adc_fsc_register(self):
        """Returns the content of the full-scale calibration register.

        Returns:
            Content of the full-scale calibration register.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=set_tsm_adc_fsc_register)
    def set_tsm_adc_fsc_register(self, cargo1: int = 0, cargo2: int = 0):
        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_command_latch)
    def tsm_adc_command_latch(self, cargo1: int = 0, cargo2: int = 0):
        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_command)
    def tsm_adc_command(self, cargo1: int = 0, cargo2: int = 0):
        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_calibration)
    def tsm_adc_calibration(self, cargo1: int = 0, cargo2: int = 0):
        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_value_xx_currentn)
    def tsm_adc_value_xx_currentn(self, probe: int = 1):
        """Returns the negative current to polarise the given thermistor.

        Args:
            probe (int): Thermistor identifier.

        Returns:
            Negative current to polarise the given thermistor.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_value_xx_biasn)
    def tsm_adc_value_xx_biasn(self, probe: int = 1):
        """Returns the voltage measured on the given thermistor biased with negative current.

        Args:
            probe (int): Thermistor identifier.

        Returns:
            Voltage on the thermistor biased with the negative current.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_value_xx_currentp)
    def tsm_adc_value_xx_currentp(self, probe: int = 1):
        """Returns the positive current to polarise the given thermistor.

        Args:
            probe (int): Thermistor identifier.

        Returns:
            Positive current to polarise the given thermistor.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_adc_value_xx_biasp)
    def tsm_adc_value_xx_biasp(self, probe: int = 1):
        """Returns the voltage measured on the given thermistor biased with positive current.

        Args:
            probe (int): Thermistor identifier.

        Returns:
            Voltage on the thermistor biased with the positive current.
        """

        pass

    @tsm_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=tsm_acq_counter)
    def tsm_acq_counter(self):
        """Reads the number of ADC measurement sequences that have been made.

        Returns:
            Number of ADC measurement sequences that have been made.
        """

        pass

    # HK commands

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=vhk_psu_vmotor)
    def vhk_psu_vmotor(self):
        """Returns the HK PSU motor voltage value.

        Returns:
            HK PSU motor voltage value.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=vhk_psu_vhi)
    def vhk_psu_vhi(self):
        """Returns the HK PSU high voltage value.

        Returns:
            HK PSU motor high voltage value.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=vhk_psu_vlow)
    def vhk_psu_vlow(self):
        """Returns the HK PSU low voltage value.

        Returns:
            HK PSU motor low voltage value.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=vhk_psu_vmedp)
    def vhk_psu_vmedp(self):
        """Returns the HK PSU medium positive voltage value.

        Returns:
            HK PSU medium positive voltage value.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=vhk_psu_vmedn)
    def vhk_psu_vmedn(self):
        """Returns the HK PSU medium negative voltage value.

        Returns:
            HK PSU medium negative voltage value.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ihk_psu_vmedn)
    def ihk_psu_vmedn(self):
        """Returns the HK PSU medium negative current value.

        Returns:
            HK PSU medium negative current value.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ihk_psu_vmedp)
    def ihk_psu_vmedp(self):
        """Returns the HK PSU medium positive current value.

        Returns:
            HK PSU medium positive current value.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ihk_psu_vlow)
    def ihk_psu_vlow(self):
        """Returns the HK PSU low current value.

        Returns:
            HK PSU low current value.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ihk_psu_vhi)
    def ihk_psu_vhi(self):
        """Returns the HK PSU high current value.

        Returns:
            HK PSU high current value.
        """

        pass

    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=ihk_psu_vmotor)
    def ihk_psu_vmotor(self):
        """Returns the HK PSU motor current value.

        Returns:
            HK PSU motor current value.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_psu_first)
    def thk_psu_first(self):
        """Returns the HK PSU temperature zone 1.

        Returns:
            HK PSU temperature zone 1.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_m2md_first)
    def thk_m2md_first(self):
        """Returns the HK M2MD temperature zone 1.

        Returns:
            HK M2MD temperature zone 1.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_psu_second)
    def thk_psu_second(self):
        """Returns the HK M2MD temperature zone 2.

        Returns:
            HK M2MD temperature zone 2.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_m2md_second)
    def thk_m2md_second(self):
        """Returns the HK M2MD temperature zone 2.

        Returns:
            HK M2MD temperature zone 2.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_cts_q1)
    def thk_cts_q1(self):
        """Returns the HK CTS temperature first quarter.

        Returns:
            HK CTS temperature first quarter.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_cts_q2)
    def thk_cts_q2(self):
        """Returns the HK CTS temperature second quarter.

        Returns:
            HK CTS temperature second quarter.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_cts_q3)
    def thk_cts_q3(self):
        """Returns the HK CTS temperature third quarter.

        Returns:
            HK CTS temperature third quarter.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_cts_q4)
    def thk_cts_q4(self):
        """Returns the HK CTS temperature fourth quarter.

        Returns:
            HK CTS temperature fourth quarter.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_cts_fpga)
    def thk_cts_fpga(self):
        """Returns the HK CTS temperature FPGA.

        Returns:
            HK CTS temperature FPGA.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=thk_cts_ads1282)
    def thk_cts_ads1282(self):
        """Returns the HK CTS temperature ADS1282.

        Returns:
            HK CTS temperature ADS1282.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=vhk_ths_ret)
    def vhk_ths_ret(self):
        """Returns the HK CTS thermistors return voltage.

        Returns:
            HK CTS thermistors return voltage.
        """

        pass

    @hk_cmd()
    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string_func=hk_acq_counter)
    def hk_acq_counter(self):
        """Returns the running counter that indicates the number of HK measurement sequences that have been made.

        Returns:
            Free running counter that indicates the number of HK measurement sequences that have been made.
        """

        pass


class TcuController(TcuInterface, DynamicCommandMixin):
    def __init__(self):
        """Initialisation of an Ariel TCU controller."""

        super().__init__()
        self.transport = self.tcu = TcuDeviceInterface()

    # noinspection PyMethodMayBeStatic
    def is_simulator(self):
        return False

    def is_connected(self):
        """Checks whether the serial port to the TCU Arduino is open.

        Returns:
            True if the serial port to the TCU Arduino is open; False otherwise.
        """

        return self.tcu.is_connected()

    def connect(self):
        """Opens the serial port to the TCU Arduino.

        Raises:
            TcuError: When the serial port could not be opened.
        """

        self.tcu.connect()

    def disconnect(self):
        """Closes the serial port to the TCU Arduino.

        Raises:
            TcuError: When the serial port could not be closed.
        """

        self.tcu.disconnect()

    def reconnect(self):
        """Re-connects to the Ariel TCU Arduino."""

        self.tcu.reconnect()


class TcuSimulator(TcuInterface):
    VHK_PSU_MOTOR = 0x0080
    VHK_PSU_VHI = 0x0280
    VHK_PSU_VLOW = 0x0480
    VHK_PSU_VMEDP = 0x0680
    VHK_PSU_VMEDN = 0x0880
    IHK_PSU_VMEDN = 0x0A80
    IHK_PSU_VMEDP = 0x0A80
    IHK_PSU_VLOW = 0x0A80
    IHK_PSU_VHI = 0x0A80
    IHK_PSU_VMOTOR = 0x0A80
    THK_PSU_FIRST = 0x0A80
    THK_M2MD_FIRST = 0x0A80
    THK_PSU_SECOND = 0x0C80
    THK_M2MD_SECOND = 0x0C80
    THK_CTS_Q1 = 0x0A80
    THK_CTS_Q2 = 0x0C80
    THK_CTS_Q3 = 0x0C80
    THK_CTS_Q4 = 0x0C80
    THK_CTS_FPGA = 0x0A80
    THK_CTS_ADS1282 = 0x0C80
    VHK_THS_RET = 0x0E80
    HK_ACQ_COUNTER = 0x002A

    def __init__(self):
        """Initialisation of an Ariel TCU simulator."""

        super().__init__()

        self._is_connected = True

        self.tcu_mode = TcuMode.IDLE.value
        self.restart_links_period = 0
        self.acq_curr_off_corr_list = [None, 0, 0, 0]
        self.acq_curr_gain_corr_list = [None, 0, 0, 0]

        self.acq_axis_a_curr_read_list = [0, 1, 2, 3]
        self.acq_axis_b_curr_read_list = [0, 4, 5, 6]

        self.prof_gen_axis_step_list = [None, 7, 8, 9]
        self.prof_gen_axis_speed_list = [None, 10, 11, 12]
        self.prof_gen_axis_state_start_list = [None, 13, 14, 14]

    # noinspection PyMethodMayBeStatic
    def is_simulator(self):
        return True

    # noinspection PyMethodMayBeStatic
    def is_connected(self):
        return self._is_connected

    def connect(self):
        self._is_connected = True

    def disconnect(self):
        self._is_connected = False

    def reconnect(self):
        self._is_connected = True

    def tcu_firmware_id(self):
        return "TCU Simulator"

    def get_tcu_mode(self):
        return self.tcu_mode

    def set_tcu_mode(self, tcu_mode: TcuMode | int = TcuMode.IDLE):
        if isinstance(tcu_mode, TcuMode):
            self.tcu_mode = tcu_mode.value

        elif isinstance(tcu_mode, int):
            self.tcu_mode = tcu_mode

    def set_restart_links_period(self, link_period: int = 0xFFFF):
        self.restart_links_period = link_period

    def get_restart_links_period(self):
        return self.restart_links_period

    def get_acq_curr_off_corr(self, axis: CommandAddress | str | int):
        return self.acq_curr_off_corr_list[int(axis)]

    def set_acq_curr_off_corr(self, axis: CommandAddress | str | int, cargo2: int = 0x03FB):
        self.acq_curr_off_corr_list[int(axis)] = cargo2

    def get_acq_curr_gain_corr(self, axis: CommandAddress | str | int):
        return self.acq_curr_gain_corr_list[int(axis)]

    def set_acq_curr_gain_corr(self, axis: CommandAddress | str | int, cargo2: int = 0x074C):
        self.acq_curr_gain_corr_list[int(axis)] = cargo2

    def acq_axis_a_curr_read(self, axis: CommandAddress | str | int):
        return self.acq_axis_a_curr_read_list[int(axis)]

    def acq_axis_b_curr_read(self, axis: CommandAddress | str | int):
        return self.acq_axis_b_curr_read_list[int(axis)]

    def get_prof_gen_axis_step(self, axis: CommandAddress | str | int):
        return self.prof_gen_axis_step_list[int(axis)]

    def set_prof_gen_axis_step(self, axis: CommandAddress | str | int, cargo2: int = 0x0480):
        self.prof_gen_axis_step_list[int(axis)] = cargo2

    def get_prof_gen_axis_speed(self, axis: CommandAddress | str | int):
        return self.prof_gen_axis_speed_list[int(axis)]

    def set_prof_gen_axis_speed(self, axis: CommandAddress | str | int, cargo2: int = 0x1777):
        self.prof_gen_axis_speed_list[int(axis)] = cargo2

    def get_prof_gen_axis_state_start(self, axis: CommandAddress | str | int):
        return self.prof_gen_axis_state_start_list[int(axis)]

    def set_prof_gen_axis_state_start(self, axis: CommandAddress | str | int, cargo2: int = 0):
        self.prof_gen_axis_state_start_list[int(axis)] = cargo2

    def vhk_psu_vmotor(self):
        return TcuSimulator.VHK_PSU_MOTOR

    def vhk_psu_vhi(self):
        return TcuSimulator.VHK_PSU_VHI

    def vhk_psu_vlow(self):
        return TcuSimulator.VHK_PSU_VLOW

    def vhk_psu_vmedp(self):
        return TcuSimulator.VHK_PSU_VMEDP

    def vhk_psu_vmedn(self):
        return TcuSimulator.VHK_PSU_VMEDN

    def ihk_psu_vmedn(self):
        return TcuSimulator.IHK_PSU_VMEDN

    def ihk_psu_vmedp(self):
        return TcuSimulator.IHK_PSU_VMEDP

    def ihk_psu_vlow(self):
        return TcuSimulator.IHK_PSU_VLOW

    def ihk_psu_vhi(self):
        return TcuSimulator.IHK_PSU_VHI

    def ihk_psu_vmotor(self):
        return TcuSimulator.IHK_PSU_VMOTOR

    def thk_psu_first(self):
        return TcuSimulator.THK_PSU_FIRST

    def thk_m2md_first(self):
        return TcuSimulator.THK_M2MD_FIRST

    def thk_psu_second(self):
        return TcuSimulator.THK_PSU_SECOND

    def thk_m2md_second(self):
        return TcuSimulator.THK_M2MD_SECOND

    def thk_cts_q1(self):
        return TcuSimulator.THK_CTS_Q1

    def thk_cts_q2(self):
        return TcuSimulator.THK_CTS_Q2

    def thk_cts_q3(self):
        return TcuSimulator.THK_CTS_Q3

    def thk_cts_q4(self):
        return TcuSimulator.THK_CTS_Q4

    def thk_cts_fpga(self):
        return TcuSimulator.THK_CTS_FPGA

    def thk_cts_ads1282(self):
        return TcuSimulator.THK_CTS_ADS1282

    def vhk_ths_ret(self):
        return TcuSimulator.VHK_THS_RET

    def hk_acq_counter(self):
        return TcuSimulator.HK_ACQ_COUNTER


class TcuProxy(DynamicProxy, TcuInterface):
    """
    The TcuProxy class is used to connect to the TCU Control Server and send commands to the TCU Hardware Controller remotely.
    """

    def __init__(self):
        """Initialisation of a TCUProxy."""

        # Fixed ports -> Use information from settings
        # This allows use to use the TCU CS without the core services

        if COMMANDING_PORT != 0:
            super().__init__(connect_address(PROTOCOL, HOSTNAME, COMMANDING_PORT))

        # Dynamic port allocation -> Use Registry Client

        else:
            with RegistryClient() as reg:
                service = reg.discover_service(SERVICE_TYPE)

                if service:
                    protocol = service.get("protocol", "tcp")
                    hostname = service["host"]
                    port = service["port"]

                    super().__init__(connect_address(protocol, hostname, port), timeout=PROXY_TIMEOUT)

                else:
                    raise RuntimeError(f"No service registered as {SERVICE_TYPE}")


class TcuHex(TcuInterface, DynamicCommandMixin):
    def __init__(self):
        super().__init__()
        self.transport = self.tcu = TcuHexInterface()



__all__ = [
    "DEFAULT_COLLECTOR_PORT",
    "DEFAULT_PUBLISHER_PORT",
    "DEFAULT_REQUESTS_PORT",
    "STATS_INTERVAL",
    "is_notify_hub_active",
    "async_is_notify_hub_active",
]

from egse.settings import Settings

settings = Settings.load("Notify Hub")

PROCESS_NAME = settings.get("PROCESS_NAME", "nh_cs")
SERVICE_ID = settings.get("SERVICE_ID", "nh_cs_1")
SERVICE_TYPE = settings.get("SERVICE_TYPE", "NH_CS")

# Default ports that are assigned to PUSH-PULL, PUB-SUB, ROUTER-DEALER protocols of the notification hub.
# The actual ports are defined in the Settings.yaml, use local settings to change them.
DEFAULT_COLLECTOR_PORT = settings.get("COLLECTOR_PORT", 0)
DEFAULT_PUBLISHER_PORT = settings.get("PUBLISHER_PORT", 0)
DEFAULT_REQUESTS_PORT = settings.get("REQUESTS_PORT", 0)

STATS_INTERVAL = settings.get("STATS_INTERVAL", 30)
"""How often the notification hub sends statistics about active connections and processed events. Default to 30s."""


async def async_is_notify_hub_active(timeout: float = 0.5) -> bool:
    """Check if the notification hub is running and active.

    This function will send a 'health_check' request to the notification hub and
    waits for the answer. If the hub replies with `healthy` the function returns
    True, otherwise, False is returned.

    If no reply was received after the given timeout [default=0.5s] the request
    will time out and return False.

    Use this function in an asynchronous context.
    """

    from egse.notifyhub.client import AsyncNotificationHubClient  # prevent circular import

    with AsyncNotificationHubClient(request_timeout=timeout) as client:
        if not await client.health_check():
            return False
        else:
            return True


def is_notify_hub_active(timeout: float = 0.5):
    """Check if the notification hub is running and active.

    This function will send a 'health_check' request to the notification hub and
    waits for the answer. If the hub replies with `healthy` the function returns
    True, otherwise, False is returned.

    If no reply was received after the given timeout [default=0.5s] the request
    will time out and return False.
    """

    from egse.notifyhub.client import NotificationHubClient  # prevent circular import

    with NotificationHubClient(request_timeout=timeout) as client:
        if not client.health_check():
            return False
        else:
            return True



"""
Testing the Notification Hub.

What is expected to happen:

  - 10 events of type 'simple-event' are published where the 'count' is incremented from [0-9] and the 'quit' is False
  - subscription to event type 'simple-event'
  - handle is registered for the even type 'simple-event'
  - when the 'count' reaches 5, listening should be interrupted and stopped
  - four more events are published with count = [6-9]
  - a last event is published with count = 9 and 'quit' = True

You can use the script 'script_subscribe_to_notifyhub.py' to follow the events.

"""

import threading
import time

import pytest

from egse.log import logger
from egse.notifyhub import is_notify_hub_active
from egse.notifyhub.event import NotificationEvent
from egse.notifyhub.services import EventPublisher
from egse.notifyhub.services import EventSubscriber

pytestmark = pytest.mark.skipif(
    not is_notify_hub_active(), reason="The notification hub shall be running for this test."
)


def test_simple_event():
    publisher = EventPublisher()
    publisher.connect()

    is_running = True

    def handle_simple_event(event_data: dict):
        # The asserts will not fail the test, but they will appear in the output
        # of the test as ERROR when they fail.
        nonlocal is_running
        assert event_data["data"]["message"] == "a unit test for event notification"
        if "quit" in event_data["data"]:
            logger.info(f"{event_data=}")
            if event_data["data"]["count"] >= 5:
                is_running = False

    def _start_listening():
        logger.info("Starting to listen...")
        subscriber = EventSubscriber(["simple-event"])
        subscriber.connect()
        subscriber.register_handler("simple-event", handle_simple_event)

        while is_running:
            if subscriber.poll():
                subscriber.handle_event()

        subscriber.disconnect()

    thread = threading.Thread(target=_start_listening)
    thread.start()

    for idx in range(10):
        publisher.publish(
            NotificationEvent(
                event_type="simple-event",
                source_service="test-simple-event",
                data={"message": "a unit test for event notification", "count": idx, "quit": False},
            )
        )
        time.sleep(0.01)

    publisher.publish(
        NotificationEvent(
            event_type="simple-event",
            source_service="test-simple-event",
            data={"message": "a unit test for event notification", "count": 10, "quit": True},
        )
    )

    time.sleep(0.1)

    publisher.disconnect()


def test_event_retention_time_1():
    # First publish the event
    # Then, after 1s, subscribe and handle the event -> should fail!

    publisher = EventPublisher()
    publisher.connect()

    publisher.publish(
        NotificationEvent(
            event_type="retention-event",
            source_service="test-event-retention-time",
            data={"message": "a unit test for testing event retention time", "count": 0, "quit": False},
        )
    )

    logger.info("Waiting for 1s...")

    time.sleep(1.0)

    logger.info("Starting to listen...")
    subscriber = EventSubscriber(["retention-event"])
    subscriber.connect()

    if subscriber.poll():
        event_data = subscriber.handle_event(return_event_data=True)
        logger.info(f"{event_data=}")
        assert event_data["event_type"] == "retention-event"
        pytest.fail("Didn't expected retention-event because we subscribed late.")
    else:
        logger.info("As expected, no retention-event...")

    subscriber.disconnect()

    publisher.disconnect()


def test_event_retention_time_2():
    # This is the scenario we will use in the control servers, e.g. the storage
    # manager, to react on new_setup events.

    # First subscribe and poll
    # Then, after 10s, publish the event.
    # The event should be returned, not handled.

    publisher = EventPublisher()
    publisher.connect()

    is_running = True

    # In this scenario, this function is not called. When return the event_data
    # instead of calling the registered handler.
    def report_event(event_data: dict):
        logger.info(f"++ {event_data=}")

    def listen_for_events():
        logger.info("Starting to listen...")
        subscriber = EventSubscriber(["retention-event"])
        subscriber.connect()
        subscriber.register_handler("simple-event", report_event)

        while is_running:
            if subscriber.poll():
                event_data = subscriber.handle_event(return_event_data=True)  # don't call registered handler
                logger.info(f"-- {event_data=}")
                assert event_data["event_type"] == "retention-event"
            else:
                logger.info("Waiting for retention-event...")

        subscriber.disconnect()

    thread = threading.Thread(target=listen_for_events)
    thread.start()

    time.sleep(10.0)

    publisher.publish(
        NotificationEvent(
            event_type="retention-event",
            source_service="test-event-retention-time",
            data={"message": "a unit test for testing event retention time", "count": 0, "quit": False},
        )
    )

    publisher.disconnect()

    is_running = False

    time.sleep(1.0)


def test_event_publisher(caplog):
    # NOTE: Make sure logging through the egse_logger is enabled for DEBUG

    caplog.set_level("DEBUG", "egse")

    # This test is to check if the message
    with EventPublisher() as pub:
        pub.publish(NotificationEvent(event_type="new_setup", source_service="cm_cs", data={"setup_id": "0001234"}))

    assert "Published" in caplog.text



from egse.env import get_conf_data_location
from egse.env import get_data_storage_location
from egse.env import get_log_file_location


def test_data_storage_layout(default_env):
    data_root = default_env.data_root

    assert get_data_storage_location() == data_root
    assert get_conf_data_location() == f"{data_root}/conf"
    assert get_log_file_location() == f"{data_root}/log"



from __future__ import annotations

import asyncio
import json
import threading
import uuid
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Union

import zmq
import zmq.asyncio

from egse.env import bool_env
from egse.log import logging
from egse.registry import DEFAULT_RS_HB_PORT
from egse.registry import DEFAULT_RS_PUB_PORT
from egse.registry import DEFAULT_RS_REQ_PORT
from egse.registry import MessageType
from egse.system import do_every

VERBOSE_DEBUG = bool_env("VERBOSE_DEBUG")

REQUEST_TIMEOUT = 0.5
"""time to wait for a request-reply [seconds]. For responsive message handling."""
EVENT_POLL_TIMEOUT = 0.5
"""time to wait for while listening for events [seconds]."""
TASK_COMPLETION_TIMEOUT = 5.0
"""time to wait for Tasks to complete [seconds]."""
HEALTH_CHECK_TIMEOUT = 2.0
"""Health checks shouldn't take too long. Recommended: 1-5s."""
HEART_BEAT_TIMEOUT = 1.0
"""Heart beat should fail fast to detect network issues. Recommended: 0.5-2s."""
HEART_BEAT_RECONNECT = 1.0
"""Time to wait between a disconnect an a connect to the heartbeat socket."""
THREAD_JOIN_TIMEOUT = 10.0
"""Joining threads should be longer, but the key is balancing graceful cleanup with reasonable shutdown times."""
PROCESS_SHUTDOWN_TIMEOUT = 10.0
"""Waiting for a process shutdown confirmation message."""


class RegistryClient:
    """
    Synchronous client for the service registry.
    """

    def __init__(
        self,
        registry_req_endpoint: str | None = None,
        registry_sub_endpoint: str | None = None,
        registry_hb_endpoint: str | None = None,
        timeout: float = REQUEST_TIMEOUT,
        client_id: str = "registry-client",
    ):
        """
        Initialize the async registry client.

        Args:
            registry_req_endpoint: ZeroMQ endpoint for REQ-REP socket, defaults to DEFAULT_RS_REQ_PORT on localhost.
            registry_sub_endpoint: ZeroMQ endpoint for SUB socket, defaults to DEFAULT_RS_PUB_PORT on localhost.
            registry_hb_endpoint: ZeroMQ endpoint for Heartbeat socket, defaults to DEFAULT_RS_HB_PORT on localhost.
            timeout: Timeout for requests in seconds, defaults to 0.5s.
            client_id: client identification, default='registry-client'
        """
        self.registry_req_endpoint = registry_req_endpoint or f"tcp://localhost:{DEFAULT_RS_REQ_PORT}"
        self.registry_sub_endpoint = registry_sub_endpoint or f"tcp://localhost:{DEFAULT_RS_PUB_PORT}"
        self.registry_hb_endpoint = registry_hb_endpoint or f"tcp://localhost:{DEFAULT_RS_HB_PORT}"

        self.timeout = timeout
        self.timeout_ms = int(timeout * 1000)
        self.logger = logging.getLogger("egse.registry.client")

        # Service state
        self._service_id = None
        self._service_info = None
        self._ttl = None

        self._hb_stop_event = None
        self._hb_thread: threading.Thread | None = None

        self._client_id = f"{client_id}-{uuid.uuid4()}".encode()

        self.context: zmq.Context = zmq.Context.instance()

        self.req_socket: zmq.Socket | None = None
        self.sub_socket: zmq.Socket | None = None
        self.hb_socket: zmq.Socket | None = None

        self.poller = zmq.Poller()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        if VERBOSE_DEBUG:
            self.logger.debug("Connecting to service registry...")

        # REQ socket for request-reply pattern
        self.req_socket = self.context.socket(zmq.DEALER)
        self.req_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
        self.req_socket.setsockopt(zmq.IDENTITY, self._client_id)
        self.req_socket.connect(self.registry_req_endpoint)

        # SUB socket for receiving events
        self.sub_socket = self.context.socket(zmq.SUB)
        self.sub_socket.connect(self.registry_sub_endpoint)
        # Default to receiving all events
        self.sub_socket.setsockopt_string(zmq.SUBSCRIBE, "")

        self.poller.register(self.req_socket, zmq.POLLIN)
        self.poller.register(self.sub_socket, zmq.POLLIN)

    def _connect_hb_socket(self):
        self.hb_socket = self.context.socket(zmq.DEALER)
        self.hb_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
        self.hb_socket.setsockopt(zmq.IDENTITY, self._client_id)
        self.hb_socket.connect(self.registry_hb_endpoint)

    def disconnect(self):
        if VERBOSE_DEBUG:
            self.logger.debug("Disconnecting from service registry...")

        if self.req_socket:
            self.poller.unregister(self.req_socket)
            self.req_socket.close(linger=0)
        self.req_socket = None

        if self.sub_socket:
            self.poller.unregister(self.sub_socket)
            self.sub_socket.close(linger=0)
        self.sub_socket = None

    def _disconnect_hb_socket(self):
        if self.hb_socket:
            self.hb_socket.setsockopt(zmq.LINGER, 0)
            self.hb_socket.close()
        self.hb_socket = None

    def _send_request(self, msg_type: MessageType, request: dict[str, Any], timeout: float = None) -> dict[str, Any]:
        """
        Send a request to the registry and get the response.

        Args:
            msg_type: the type of message and reply
            request: The request to send
            timeout: The number of seconds to wait, if None, instance variable is used.

        Returns:
            The response from the registry.
        """
        timeout_ms = int(timeout * 1000) if timeout else self.timeout_ms

        try:
            if VERBOSE_DEBUG:
                self.logger.debug(f"Sending request: {request}")

            self.req_socket.send_multipart([msg_type.value, json.dumps(request).encode()])

            if msg_type == MessageType.REQUEST_NO_REPLY:
                return {"success": True}

            if self.poller.poll(timeout=timeout_ms):
                message_parts = self.req_socket.recv_multipart()

                if len(message_parts) >= 2:
                    message_type = MessageType(message_parts[0])
                    message_data = message_parts[1]

                    if message_type == MessageType.RESPONSE:
                        response = json.loads(message_data)
                        if VERBOSE_DEBUG:
                            self.logger.debug(f"Received response: {response}")
                        return response
                    else:
                        return {
                            "success": False,
                            "error": f"unexpected MessageType received: {message_type.name}, {message_data = }",
                        }
                else:
                    return {
                        "success": False,
                        "error": f"not enough parts received: {len(message_parts)}",
                        "data": message_parts,
                    }
            else:
                self.logger.error(f"Request timed out after {self.timeout:.2f}s")
                return {"success": False, "error": "Request timed out"}
        except zmq.ZMQError as exc:
            self.logger.error(f"ZMQ error: {exc}", exc_info=True)
            return {"success": False, "error": str(exc)}
        except Exception as exc:
            self.logger.error(f"Error sending request: {exc}", exc_info=True)
            return {"success": False, "error": str(exc)}

    def _send_heartbeat(self, request: dict[str, Any]) -> dict[str, Any]:
        """
        Send a heartbeat to the registry and get the response.

        Args:
            request: The request to send

        Returns:
            The response from the registry.
        """

        try:
            if VERBOSE_DEBUG:
                self.logger.debug(f"Sending heartbeat request: {request}")

            self.hb_socket.send_string(json.dumps(request))

            ready_sockets, _, _ = zmq.select([self.hb_socket], [], [], timeout=HEART_BEAT_TIMEOUT)

            if ready_sockets:
                response_json = self.hb_socket.recv_string()
                response = json.loads(response_json)
                if VERBOSE_DEBUG:
                    self.logger.debug(f"Received response: {response}")
                return response
            else:
                self.logger.error(f"Request timed out after {HEART_BEAT_TIMEOUT:.2f}s")
                return {"success": False, "error": "Request timed out"}

        except zmq.ZMQError as exc:
            self.logger.error(f"ZMQ error: {exc}", exc_info=True)
            return {"success": False, "error": str(exc)}
        except Exception as exc:
            self.logger.error(f"Error sending request: {exc}", exc_info=True)
            return {"success": False, "error": str(exc)}

    def register(
        self,
        name: str,
        host: str,
        port: int,
        service_type: str | None = None,
        metadata: dict[str, Any] | None = None,
        ttl: int = 30,
    ) -> str | None:
        """
        Register this service with the registry.

        Args:
            name: Service name
            host: Service host/IP
            port: Service port
            service_type: Service type (for discovery)
            metadata: Additional service metadata
            ttl: Time-to-live in seconds

        Returns:
            The service ID if successful, None otherwise
        """
        # Prepare service info
        service_info: dict[str, str | int | dict | list] = {"name": name, "host": host, "port": port}

        # Add optional fields
        if service_type:
            service_info["type"] = service_type

        if metadata:
            service_info["metadata"] = metadata

        # Prepare tags for easier discovery
        tags = []
        if service_type:
            tags.append(service_type)
        service_info["tags"] = tags

        # Send registration request
        request = {"action": "register", "service_info": service_info, "ttl": ttl}

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get("success"):
            # Store service information for later use
            self._service_id = response.get("service_id")
            self._service_info = service_info
            self._ttl = ttl

            self.logger.info(f"Service registered with ID: {self._service_id}")
            return self._service_id
        else:
            self.logger.error(f"Failed to register service: {response.get('error')}")
            return None

    def deregister(self, service_id: str = None) -> bool:
        """
        Deregister this service from the registry.

        Returns:
            True if successful, False otherwise.
        """

        service_id = service_id or self._service_id

        if not service_id:
            self.logger.warning("Cannot deregister: no service is registered")
            return False

        request = {"action": "deregister", "service_id": service_id}

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get("success"):
            self.logger.info(f"Service deregistered: {service_id}")
            self._service_id = None
            self._service_info = None
            self._ttl = None
            return True
        else:
            self.logger.error(f"Failed to deregister service: {response.get('error')}")
            return False

    def reregister(self) -> str | None:
        if not self._service_id:
            self.logger.warning("Cannot reregister: no service is registered")
            return None

        if not self._service_info:
            self.logger.warning(
                "Cannot reregister: no service info was saved by this registry client or service already deregistered."
            )
            return None

        return self.register(
            name=self._service_info["name"],
            host=self._service_info["host"],
            port=self._service_info["port"],
            service_type=self._service_info["type"],
            metadata=self._service_info["metadata"],
        )

    def discover_service(self, service_type: str) -> dict[str, Any] | None:
        """
        Discover a service of the specified type. The service is guaranteed to be healthy at the time of discovery.

        The returned information contains:

        - name: the name of the service
        - host: the ip address or hostname of the service
        - port: the port number for requests to the microservice

        Args:
            service_type: Type of service to discover

        Returns:
            Service information if found, None otherwise
        """
        request = {"action": "discover", "service_type": service_type}

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if VERBOSE_DEBUG:
            self.logger.debug(f"{response = }")

        if response.get("success"):
            service = response.get("service")
            return service
        else:
            if VERBOSE_DEBUG:
                self.logger.debug(f"Service discovery failed: {response.get('error')}")
            return None

    def get_service(self, service_id: str | None = None) -> dict[str, Any] | None:
        """
        Get information about a specific service. When no service_id is given,
        the service_id known to this client will be used.

        Args:
            service_id: ID of the service to get [default=None]

        Returns:
            Service information if found, None otherwise.
        """
        service_id = service_id or self._service_id

        if not service_id:
            self.logger.warning("Cannot get service: no service id is given or known.")
            return None

        request = {"action": "get", "service_id": service_id}

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get("success"):
            service = response.get("service")
            return service
        else:
            self.logger.warning(f"Get service failed: {response.get('error')}")
            return None

    def list_services(self, service_type: str | None = None) -> list[dict[str, Any]]:
        """
        List all registered services, optionally filtered by type.

        Args:
            service_type: Type of services to list

        Returns:
            List of service information.
        """
        request = {"action": "list", "service_type": service_type}

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get("success"):
            services = response.get("services", [])
            return services
        else:
            self.logger.warning(f"List services failed: {response.get('error')}")
            return []

    def get_endpoint(self, service_type) -> str | None:
        """Returns the endpoint for the given service type or None if the service_type is not registered."""
        service = self.discover_service(service_type)

        if service:
            protocol = service.get("protocol", "tcp")
            hostname = service["host"]
            port = service["port"]

            return f"{protocol}://{hostname}:{port}"
        else:
            return None

    def health_check(self) -> bool:
        """
        Check if the registry server is healthy.

        Returns:
            True if healthy, False otherwise
        """
        request = {"action": "health"}

        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request, timeout=HEALTH_CHECK_TIMEOUT)
        return response.get("success", False)

    def terminate_registry_server(self) -> bool:
        """
        Send a terminate request to the service registry server. Returns True when successful.
        """
        request = {"action": "terminate"}
        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request, timeout=PROCESS_SHUTDOWN_TIMEOUT)
        return response.get("success", False)

    def server_status(self) -> dict[str, Any]:
        """
        Requests the status information from the service registry.
        """
        request = {
            "action": "info",
        }
        response = self._send_request(MessageType.REQUEST_WITH_REPLY, request)
        return response

    def start_heartbeat(self, interval: int | None = None) -> None:
        if not self._service_id:
            self.logger.warning("Cannot start heartbeat: no service is registered")
            return None

        # Cancel an existing heartbeat thread if present
        self.stop_heartbeat()

        # If interval not specified, use 1/3 of TTL
        if interval is None:
            interval = max(1, self._ttl // 3)

        def send_heartbeat():
            try:
                request = {"action": "renew", "service_id": self._service_id}

                response = self._send_heartbeat(request)

                if not response.get("success", False):
                    self.logger.warning(f"Heartbeat failed: {response.get('error')}")

                    # Do a health check
                    if not self.health_check():
                        self.logger.warning("Heartbeat failed: ServiceRegistry not responding.")
                        return
                    else:
                        self.logger.info("Heartbeat failed, but health check succeeded, reregistering...")
                        self.reregister()

                else:
                    if VERBOSE_DEBUG:
                        self.logger.debug(f"Heartbeat succeeded: {response.get('message')}")

            except Exception as exc:
                self.logger.error(f"Error sending heartbeat: {exc}")

        self._hb_stop_event = threading.Event()
        self._hb_thread = threading.Thread(
            target=do_every,
            args=(interval, send_heartbeat),
            kwargs={
                "stop_event": self._hb_stop_event,
                "setup_func": self._connect_hb_socket,
                "teardown_func": self._disconnect_hb_socket,
            },
        )
        self._hb_thread.daemon = True
        self._hb_thread.start()
        return None

    def stop_heartbeat(self) -> None:
        if self._hb_stop_event:
            self._hb_stop_event.set()
        if self._hb_thread:
            if self._hb_thread.is_alive():
                self.logger.debug(f"Heartbeat thread is alive, joining with timeout={THREAD_JOIN_TIMEOUT}s")
                self._hb_thread.join(timeout=THREAD_JOIN_TIMEOUT)
            else:
                self.logger.debug("Heartbeat thread is not alive.")
                self._hb_thread = None
        else:
            self.logger.debug("No heartbeat thread defined.")

    def close(self) -> None:
        """Clean up resources."""
        try:
            if hasattr(self, "req_socket") and self.req_socket:
                self.req_socket.close()

            if hasattr(self, "sub_socket") and self.sub_socket:
                self.sub_socket.close()

            if hasattr(self, "context") and self.context:
                self.context.term()
        except Exception as exc:
            self.logger.error(f"Error during cleanup: {exc}")


class AsyncRegistryClient:
    """
    Asynchronous client for interacting with the ZeroMQ-based service registry.

    This class uses asyncio and the ZeroMQ asyncio API for non-blocking operations.
    """

    def __init__(
        self,
        registry_req_endpoint: str = None,
        registry_sub_endpoint: str = None,
        registry_hb_endpoint: str = None,
        timeout: float = REQUEST_TIMEOUT,
        client_id: str = "registry-client",
    ):
        """
        Initialize the async registry client.

        Args:
            registry_req_endpoint: ZeroMQ endpoint for REQ-REP socket, defaults to DEFAULT_RS_REQ_PORT on localhost.
            registry_sub_endpoint: ZeroMQ endpoint for SUB socket, defaults to DEFAULT_RS_PUB_PORT on localhost.
            registry_hb_endpoint: ZeroMQ endpoint for heartbeat socket, defaults to DEFAULT_RS_HB_PORT on localhost.
            timeout: Timeout for requests in seconds, defaults to 0.5.
            client_id: client identification, default='registry-client'
        """
        self.registry_req_endpoint = registry_req_endpoint or f"tcp://localhost:{DEFAULT_RS_REQ_PORT}"
        self.registry_sub_endpoint = registry_sub_endpoint or f"tcp://localhost:{DEFAULT_RS_PUB_PORT}"
        self.registry_hb_endpoint = registry_hb_endpoint or f"tcp://localhost:{DEFAULT_RS_HB_PORT}"

        self.timeout = timeout
        self.timeout_ms = int(timeout * 1000)
        self.logger = logging.getLogger("egse.registry.client")

        # Service state
        self._service_id = None
        self._service_info = None
        self._ttl = None

        self._client_id = f"{client_id}-{uuid.uuid4()}".encode()

        self.context = zmq.asyncio.Context.instance()

        self._running = False
        self._tasks = set()
        self._heartbeat_task = None
        self._event_listener_task = None

        self._event_handlers = {}

        self.req_socket: zmq.asyncio.Socket | None = None
        self.sub_socket: zmq.asyncio.Socket | None = None
        self.hb_socket: zmq.asyncio.Socket | None = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        if VERBOSE_DEBUG:
            self.logger.debug("Connecting to service registry...")

        # REQ socket for request-reply pattern
        self.req_socket = self.context.socket(zmq.DEALER)
        self.req_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
        self.req_socket.setsockopt(zmq.IDENTITY, self._client_id)
        self.req_socket.connect(self.registry_req_endpoint)

        # SUB socket for receiving events
        self.sub_socket = self.context.socket(zmq.SUB)
        self.sub_socket.connect(self.registry_sub_endpoint)
        # Default to receiving all events
        self.sub_socket.setsockopt_string(zmq.SUBSCRIBE, "")

    def _connect_hb_socket(self):
        self.hb_socket = self.context.socket(zmq.DEALER)
        self.hb_socket.setsockopt(zmq.LINGER, 0)  # Don't wait for unsent messages on close()
        self.hb_socket.setsockopt(zmq.IDENTITY, self._client_id)
        self.hb_socket.connect(self.registry_hb_endpoint)

    def disconnect(self):
        if VERBOSE_DEBUG:
            self.logger.debug("Disconnecting from service registry...")

        if self.req_socket:
            self.req_socket.close(linger=0)
        self.req_socket = None

        if self.sub_socket:
            self.sub_socket.close(linger=0)
        self.sub_socket = None

    def _disconnect_hb_socket(self):
        if self.hb_socket:
            self.hb_socket.setsockopt(zmq.LINGER, 0)
            self.hb_socket.close()
        self.hb_socket = None

    async def _send_request(
        self, msg_type: MessageType, request: dict[str, Any], timeout: float = None
    ) -> dict[str, Any]:
        """
        Send a request to the registry and get the response.

        Args:
            msg_type: the type of message and reply
            request: The request to send to the service registry server.
            timeout: The number of seconds to wait, if None, instance variable is used.

        Returns:
            The response from the registry as a dictionary.
        """

        assert self.req_socket is not None, "REQ socket is not connected, cannot send request."

        timeout = timeout or self.timeout
        try:
            if VERBOSE_DEBUG:
                self.logger.debug(f"Sending request: {request}")

            await self.req_socket.send_multipart([msg_type.value, json.dumps(request).encode()])

            try:
                message_parts = await asyncio.wait_for(self.req_socket.recv_multipart(), timeout=timeout)

                if len(message_parts) >= 2:
                    message_type = MessageType(message_parts[0])
                    message_data = message_parts[1]

                    if message_type == MessageType.RESPONSE:
                        response = json.loads(message_data)
                        if VERBOSE_DEBUG:
                            self.logger.debug(f"Received response: {response}")
                        return response
                    else:
                        return {
                            "success": False,
                            "error": f"unexpected MessageType received: {message_type.name}, {message_data = }",
                        }
                else:
                    return {
                        "success": False,
                        "error": f"not enough parts received: {len(message_parts)}",
                        "data": message_parts,
                    }
            except asyncio.TimeoutError:
                self.logger.error(f"Request timed out after {timeout:.2f}s")
                return {"success": False, "error": "Request timed out"}
        except zmq.ZMQError as exc:
            self.logger.error(f"ZMQ error: {exc}", exc_info=True)
            return {"success": False, "error": str(exc)}
        except Exception as exc:
            self.logger.error(f"Error sending request: {exc}", exc_info=True)
            return {"success": False, "error": str(exc)}

    async def _send_heartbeat(self, request: dict[str, Any]) -> dict[str, Any]:
        """
        Send a heartbreak request to the registry and get the response.

        Args:
            request: The request to send to the service registry server.

        Returns:
            The response from the registry as a dictionary.
        """

        assert self.hb_socket is not None, "HB socket is not connected, cannot send heartbeat request."

        try:
            if VERBOSE_DEBUG:
                self.logger.debug(f"Sending heartbeat request: {request}")

            await self.hb_socket.send_string(json.dumps(request))

            try:
                response_json = await asyncio.wait_for(self.hb_socket.recv_string(), timeout=HEART_BEAT_TIMEOUT)
                response = json.loads(response_json)
                if VERBOSE_DEBUG:
                    self.logger.debug(f"Received heartbeat response: {response = }")
                return response
            except asyncio.TimeoutError:
                self.logger.error(f"Heartbeat request timed out after {HEART_BEAT_TIMEOUT:.2f}s")
                return {"success": False, "error": "Heartbeat request timed out"}

        except zmq.ZMQError as exc:
            self.logger.error(f"ZMQ error: {exc}", exc_info=True)
            return {"success": False, "error": str(exc)}
        except Exception as exc:
            self.logger.error(f"Error sending heartbeat request: {exc}", exc_info=True)
            return {"success": False, "error": str(exc)}

    async def register(
        self,
        name: str,
        host: str,
        port: int,
        service_type: str | None = None,
        metadata: dict[str, Any] | None = None,
        ttl: int = 30,
    ) -> str | None:
        """
        Register this service with the registry.

        Args:
            name: Service name
            host: Service host/IP
            port: Service port
            service_type: Service type (for discovery)
            metadata: Additional service metadata
            ttl: Time-to-live in seconds

        Returns:
            The service ID if successful, None otherwise
        """
        # Prepare service info
        service_info: dict[str, Any] = {"name": name, "host": host, "port": port}

        # Add optional fields
        if service_type:
            service_info["type"] = service_type

        if metadata:
            service_info["metadata"] = metadata

        # Prepare tags for easier discovery
        tags = []
        if service_type:
            tags.append(service_type)
        service_info["tags"] = tags

        # Send registration request
        request = {"action": "register", "service_info": service_info, "ttl": ttl}

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get("success"):
            # Store service information for later use
            self._service_id = response.get("service_id")
            self._service_info = service_info
            self._ttl = ttl

            self.logger.info(f"Service registered with ID: {self._service_id}")
            return self._service_id
        else:
            self.logger.error(f"Failed to register service: {response.get('error')}")
            return None

    async def deregister(self, service_id: str | None = None) -> bool:
        """
        Deregister this service or the service with service_id from the registry.

        When you register and deregister with the same client instance, you don't have to provide
        the service_id.

        NOTE:
            When a registry client is running and sending out heartbeats, it will reregister if it finds the
            registry server to be healthy. If you want to permanently deregister the service, you will need
            to stop the registry client for the service or terminate the service.

        Args:
            service_id: the service identifier that was previously handed out by the ServiceRegistry after
                registration. If not provided, the service_id that was saved when registering is used.

        Returns:
            True if successful, False otherwise
        """

        service_id = service_id or self._service_id

        if not service_id:
            self.logger.warning("Cannot deregister: no service is registered")
            return False

        request = {"action": "deregister", "service_id": service_id}

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get("success"):
            self.logger.info(f"Service deregistered: {service_id}")
            self._service_id = None
            self._service_info = None
            self._ttl = None
            return True
        else:
            self.logger.error(f"Failed to deregister service: {response.get('error')}")
            return False

    async def reregister(self) -> str | None:
        if not self._service_id:
            self.logger.warning("Cannot reregister: no service is registered")
            return None

        if not self._service_info:
            self.logger.warning(
                "Cannot reregister: no service info was saved by this registry client or service already deregistered."
            )
            return None

        return await self.register(
            name=self._service_info["name"],
            host=self._service_info["host"],
            port=self._service_info["port"],
            service_type=self._service_info["type"],
            metadata=self._service_info["metadata"],
        )

    async def start_heartbeat(self, interval: int | None = None) -> asyncio.Task | None:
        """
        Start sending heartbeats to the registry.

        Args:
            interval: Heartbeat interval in seconds (default: 1/3 of TTL)

        Returns:
            The heartbeat task
        """
        if not self._service_id:
            self.logger.warning("Cannot start heartbeat: no service is registered")
            return None

        # Cancel existing heartbeat task if present
        await self.stop_heartbeat()

        # If interval not specified, use 1/3 of TTL
        if interval is None:
            interval = max(1, self._ttl // 3)

        self._running = True

        async def heartbeat_loop():
            self._connect_hb_socket()

            try:
                while self._running and self._service_id:
                    try:
                        request = {"action": "renew", "service_id": self._service_id}

                        response = await self._send_heartbeat(request)

                        if not response.get("success"):
                            self.logger.warning(f"Heartbeat failed: {response.get('error')}")

                            # Do a health check
                            if not await self.health_check():
                                self.logger.warning("Heartbeat failed: ServiceRegistry not responding.")
                                continue
                            else:
                                self.logger.info("Heartbeat failed, but health check succeeded, reregistering...")
                                await self.reregister()

                        else:
                            if VERBOSE_DEBUG:
                                self.logger.debug(f"Heartbeat succeeded: {response.get('message')}")

                    except Exception as exc:
                        self.logger.error(f"Error in heartbeat loop: {exc}", exc_info=True)

                    # Sleep until next heartbeat
                    await asyncio.sleep(interval)
            except asyncio.CancelledError:
                self.logger.info("Heartbeat task cancelled")
            finally:
                self._disconnect_hb_socket()

        # Start the heartbeat task

        self.logger.info(f"Starting heartbeat task with interval {interval}s")

        self._heartbeat_task = task = asyncio.create_task(heartbeat_loop())
        self._tasks.add(task)
        task.add_done_callback(lambda t: self._tasks.discard(t))

        return task

    async def stop_heartbeat(self) -> None:
        """Stop the running heartbeat task."""

        if self._heartbeat_task is None:
            if VERBOSE_DEBUG:
                self.logger.debug("Couldn't stop heartbeat, heartbeat_task is None")
            return

        self._heartbeat_task.cancel()
        try:
            await self._heartbeat_task
        except asyncio.CancelledError:
            self.logger.info("Heartbeat task cancelled")
            pass
        self._tasks.discard(self._heartbeat_task)
        self._heartbeat_task = None
        self.logger.info("Stopped heartbeat task")

    async def stop_event_listener(self) -> None:
        """Stop the running event listener task."""

        if self._event_listener_task is None:
            if VERBOSE_DEBUG:
                self.logger.debug("Couldn't stop event_listener, event_listener_task is None")
            return

        self._event_listener_task.cancel()
        try:
            await self._event_listener_task
        except asyncio.CancelledError:
            pass
        self._tasks.discard(self._event_listener_task)
        self._event_listener_task = None
        self.logger.info("Stopped event listener task")

    async def stop_all_tasks(self) -> None:
        self._running = False

        # Cancel all tasks
        for task in self._tasks:
            task.cancel()

        # Wait for tasks to complete (with timeout)
        if self._tasks:
            try:
                await asyncio.wait(self._tasks, timeout=TASK_COMPLETION_TIMEOUT)
            except asyncio.CancelledError:
                pass

        self._tasks.clear()
        self.logger.info("Stopped all background tasks")

    def on_event(self, event_type: str, handler: Callable[[dict[str, Any]], Union[None, Coroutine]]) -> None:
        """
        Register a handler for a specific event type.

        Args:
            event_type: Type of event to handle (register, deregister, expire)
            handler: Function or coroutine to call with event data
        """
        if event_type not in self._event_handlers:
            self._event_handlers[event_type] = []

            # Subscribe to this specific event type
            self.sub_socket.setsockopt_string(zmq.SUBSCRIBE, event_type)

        self._event_handlers[event_type].append(handler)
        self.logger.debug(f"Registered handler for {event_type} events")

    async def start_event_listener(self) -> asyncio.Task:
        """
        Start listening for registry events.

        Returns:
            The event listener task
        """
        # Cancel existing event listener task if present
        await self.stop_event_listener()

        self._running = True

        async def subscription_loop():
            try:
                while self._running:
                    try:
                        # Use a timeout to allow for clean shutdown
                        try:
                            message = await asyncio.wait_for(
                                self.sub_socket.recv_multipart(), timeout=EVENT_POLL_TIMEOUT
                            )
                        except asyncio.TimeoutError:
                            continue

                        # Parse the message
                        event_type_bytes, event_json_bytes = message
                        event_type = event_type_bytes.decode("utf-8")
                        event = json.loads(event_json_bytes.decode("utf-8"))

                        self.logger.debug(f"Received event: {event_type}")

                        # Call registered handlers
                        handlers = self._event_handlers.get(event_type, [])
                        for handler in handlers:
                            try:
                                # Check if handler is a coroutine function
                                if asyncio.iscoroutinefunction(handler):
                                    await handler(event["data"])
                                else:
                                    handler(event["data"])
                            except Exception as exc:
                                self.logger.error(f"Error in event handler: {exc}")
                    except zmq.ZMQError as exc:
                        self.logger.error(f"ZMQ error in event listener: {exc}")
                    except Exception as exc:
                        self.logger.error(f"Error in event listener: {exc}")
                        await asyncio.sleep(1)  # Prevent tight loop on error
            except asyncio.CancelledError:
                self.logger.info("Event listener task cancelled")

        # Start the subscription task
        self._event_listener_task = task = asyncio.create_task(subscription_loop())
        self._tasks.add(task)
        task.add_done_callback(lambda t: self._tasks.discard(t))

        self.logger.info("Started event listener task")
        return task

    async def discover_service(self, service_type: str) -> dict[str, Any] | None:
        """
        Discover a service of the specified type. The service is guaranteed to be healthy at the time of discovery.

        The returned information contains:

        - name: the name of the service
        - host: the ip address or hostname of the service
        - port: the port number for requests to the microservice

        Args:
            service_type: Type of service to discover

        Returns:
            Service information if found, None otherwise
        """

        request = {"action": "discover", "service_type": service_type}

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get("success"):
            service = response.get("service")

            return service
        else:
            self.logger.warning(f"Service discovery failed: {response.get('error')}")
            return None

    async def get_service(self, service_id: str | None = None) -> dict[str, Any] | None:
        """
        Get information about a specific service.  When no service_id is given,
        the service_id known to this client will be used.

        Args:
            service_id: ID of the service to get

        Returns:
            Service information if found, None otherwise
        """

        service_id = service_id or self._service_id

        request = {"action": "get", "service_id": service_id}

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get("success"):
            service = response.get("service")

            return service
        else:
            self.logger.warning(f"Get service failed: {response.get('error')}")
            return None

    async def list_services(self, service_type: str | None = None) -> list[dict[str, Any]]:
        """
        List all registered services, optionally filtered by type.

        Args:
            service_type: Type of services to list

        Returns:
            List of service information
        """
        request = {"action": "list", "service_type": service_type}

        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)

        if response.get("success"):
            services = response.get("services", [])

            return services
        else:
            self.logger.warning(f"List services failed: {response.get('error')}")
            return []

    async def get_endpoint(self, service_type) -> str | None:
        """Returns the endpoint for the given service type."""
        service = await self.discover_service(service_type)

        if service:
            protocol = service.get("protocol", "tcp")
            hostname = service["host"]
            port = service["port"]

            return f"{protocol}://{hostname}:{port}"
        else:
            return None

    async def health_check(self) -> bool:
        """
        Check if the registry server is healthy.

        Returns:
            True if healthy, False otherwise
        """
        request = {"action": "health"}
        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request, timeout=HEALTH_CHECK_TIMEOUT)
        return response.get("success", False)

    async def terminate_registry_server(self) -> bool:
        """
        Send a terminate request to the service registry server. Returns True when successful.
        """
        request = {"action": "terminate"}
        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request, timeout=PROCESS_SHUTDOWN_TIMEOUT)
        return response.get("success", False)

    async def server_status(self) -> dict[str, Any]:
        """
        Requests the status information from the service registry.
        """
        request = {"action": "info"}
        response = await self._send_request(MessageType.REQUEST_WITH_REPLY, request)
        return response

    async def close(self) -> None:
        """Clean up resources."""
        await self.stop_heartbeat()  # This stops all tasks

        try:
            if hasattr(self, "req_socket") and self.req_socket:
                self.req_socket.close()

            if hasattr(self, "sub_socket") and self.sub_socket:
                self.sub_socket.close()

            # We can not terminate the context, because we use a global instance, i.e. a singleton context.
            # When we try to terminate it, even after checking if it was closed, it raises an exception.
            if hasattr(self, "context") and self.context:
                if VERBOSE_DEBUG:
                    self.logger.debug(f"{self.context = !r}")
                    self.logger.debug(f"{self.context._sockets = !r}")
                # The zmq context instance is the global singleton instance.
                # Terminating it here would affect other parts of the application using zmq.
                # if not self.context.closed:
                #     self.context.term()
        except Exception as exc:
            self.logger.error(f"Error during cleanup: {exc}")


def is_service_registered(service_type: str):
    """Convenience function to check if a service is registered."""
    with RegistryClient() as reg:
        response = reg.discover_service(service_type)

    return False if response is None else True
